import importlib.metadata
import sys
from json import loads as json_loads
from os import PathLike, walk as os_walk
from os.path import relpath
from pathlib import Path
from re import search as re_search
from typing import TYPE_CHECKING, Optional

from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.logging import LOGGER

if TYPE_CHECKING:
    from mountaineer.controller import ControllerBase


class ManagedViewPath(type(Path())):  # type: ignore
    """
    Helper class to manage our view directory conventions. We choose the superclass of this
    class at runtime to conform to the OS-specific Path implementation.

    """

    # Root link to the view directory
    root_link: Optional["ManagedViewPath"]

    # Root link to the directory that stores package.json / node_modules / etc
    # Typically this is the same thing as the root_link, but in the case of plugins
    # they may be different
    package_root_link: Optional["ManagedViewPath"]

    def __new__(cls, *args, **kwargs):
        # Ensure instances are created properly by using the __new__ method of the Path class
        # path = ManagedViewPath(*args, **kwargs)
        path = super().__new__(cls, *args, **kwargs)
        path.root_link = None
        path.package_root_link = None
        return path

    @classmethod
    def from_view_root(
        cls,
        root_path: PathLike | str,
        package_root_link: PathLike
        | str
        | None
        | MountaineerUnsetValue = MountaineerUnsetValue(),
    ):
        """
        Constructor to create a ManagedViewPath from the view root
        """
        path = cls(root_path)
        path.root_link = path
        if isinstance(package_root_link, MountaineerUnsetValue):
            path.package_root_link = path
        elif package_root_link is not None:
            path.package_root_link = cls(package_root_link)
        return path

    def __truediv__(self, key):
        # Override '/' operation for when ManagedViewPath is on the left
        result = super().__truediv__(key)
        return self._inherit_root_link(result)

    def __rtruediv__(self, key):
        # Override '/' operation for when ManagedViewPath is on the right
        result = super().__rtruediv__(key)
        return ManagedViewPath(result)

    def get_root_link(self):
        """
        Get the original root link for this view path. If the current view is a root link, it will return itself.
        """
        if self.root_link is None:
            raise ValueError(
                f"Cannot get root link from a non-root linked view path: {self}"
            )
        return self.root_link

    def get_package_root_link(self):
        """
        Get the original package root link for this view path. If the current view is a root link, it will return itself.
        """
        if self.package_root_link is None:
            raise ValueError(
                f"Cannot get package root link from current view path: {self}"
            )
        return self.package_root_link

    def get_managed_code_dir(self):
        return self.get_managed_dir_common("_server")

    def get_managed_static_dir(self, tmp_build: bool = False):
        # Only root paths can have static directories
        if not self.is_root_link:
            raise ValueError(
                "Cannot get static directory from a non-root linked view path"
            )
        path = self.get_managed_dir_common("_static")
        if tmp_build:
            path = path / "tmp"
            path.mkdir(exist_ok=True)
        return path

    def get_managed_ssr_dir(self, tmp_build: bool = False):
        # Only root paths can have SSR directories
        if not self.is_root_link:
            raise ValueError(
                "Cannot get SSR directory from a non-root linked view path"
            )
        path = self.get_managed_dir_common("_ssr")
        if tmp_build:
            path = path / "tmp"
            path.mkdir(exist_ok=True)
        return path

    def get_managed_dir_common(self, managed_dir: str):
        # If the path is to a file, we want to get the parent directory
        # so that we can create the managed code directory
        # We also create the managed code directory if it doesn't exist so all subsequent
        # calls can immediately start writing to it
        path = self
        if path.is_file():
            path = path.parent
        managed_code_dir = path / managed_dir
        managed_code_dir.mkdir(exist_ok=True)
        return managed_code_dir

    def get_controller_view_path(self, controller: "ControllerBase"):
        """
        Assume all paths are specified in terms of their relative root
        """
        # Ensure we are being take relative to the root view
        if not self.is_root_link:
            raise ValueError(
                "Cannot get controller view path from a non-root linked view path"
            )

        controller_path = controller.view_path

        # If the user has specified a full path, we will use that
        if isinstance(controller_path, ManagedViewPath):
            # Merge in the current attributes, if they don't exist
            controller_path = controller_path.copy()

            if controller_path.root_link is None:
                controller_path.root_link = self.root_link
            if controller_path.package_root_link is None:
                controller_path.package_root_link = self.package_root_link
            return controller_path

        relative_path = Path(controller_path.lstrip("/"))
        return self / relative_path

    def copy(self) -> "ManagedViewPath":
        path = self.__class__(self)
        path.root_link = self.root_link
        path.package_root_link = self.package_root_link
        return path

    if sys.version_info >= (3, 12):

        def rglob(self, pattern: str, *, case_sensitive: bool | None = None):
            for path in super().rglob(pattern, case_sensitive=case_sensitive):
                yield self._inherit_root_link(path)
    else:

        def rglob(self, pattern: str):
            # Override the rglob method to return a ManagedViewPath
            for path in super().rglob(pattern):
                yield self._inherit_root_link(path)

    if sys.version_info >= (3, 12):

        def walk(self, top_down=True, on_error=None, follow_symlinks=False):
            for root, dirs, files in super().walk(
                top_down=top_down, on_error=on_error, follow_symlinks=follow_symlinks
            ):
                yield self._inherit_root_link(root), dirs, files

    else:

        def walk(self):
            for root, dirs, files in os_walk(str(self)):
                yield self._inherit_root_link(root), dirs, files

    def resolve(self, strict=False):
        return self._inherit_root_link(super().resolve())

    def absolute(self):
        return self._inherit_root_link(super().absolute())

    if sys.version_info >= (3, 12):

        def relative_to(self, __other, *_deprecated, walk_up: bool = False):
            return self._inherit_root_link(
                super().relative_to(__other, *_deprecated, walk_up=walk_up)
            )
    else:

        def relative_to(self, *other):
            return self._inherit_root_link(super().relative_to(*other))

    def with_name(self, name):
        return self._inherit_root_link(super().with_name(name))

    def with_suffix(self, suffix):
        return self._inherit_root_link(super().with_suffix(suffix))

    @property
    def is_root_link(self):
        return self == self.root_link

    @property
    def parent(self):  # type: ignore
        return self._inherit_root_link(super().parent)

    def _inherit_root_link(self, new_path: Path | str) -> "ManagedViewPath":
        managed_path = ManagedViewPath(new_path)
        managed_path.root_link = self.root_link
        managed_path.package_root_link = self.package_root_link
        return managed_path


def is_path_file(path: Path):
    if path.exists():
        return path.is_file()

    # If the file doesn't actually exist (common in unit tests), we guess the path
    LOGGER.warning(f"File {path} does not exist. Guessing file status.")

    # Only use is_file if the current path is ambiguous
    dot_components = [
        component for component in path.name.split(".") if component.strip()
    ]
    return len(dot_components) > 1


def generate_relative_import(
    current_import: Path,
    desired_import: Path,
    strip_js_extensions: bool = True,
):
    """
    Given the path of the current file and the file that should be imported, try to find
    a relative path that can be used to import the file.

    :param strip_js_extensions: This function is usually called by our JS import header
    constructors. In the ESM syntax importing other files shouldn't include the javascript
    suffix. This flag allows callers to strip the suffix from the import path.

    """
    current_import = current_import.resolve()
    desired_import = desired_import.resolve()

    # Calculate the relative path
    relative_path = relpath(
        desired_import,
        current_import.parent if is_path_file(current_import) else current_import,
    )

    # Convert to JavaScript import format
    if not relative_path.startswith("."):
        relative_path = "./" + relative_path

    # Strip the file extension
    if strip_js_extensions:
        for js_extensions in [".js", ".jsx", ".ts", ".tsx"]:
            if relative_path.endswith(js_extensions):
                relative_path = relative_path.removesuffix(js_extensions)

    return relative_path


def resolve_package_path(package_name: str):
    """
    Given a package distribution, returns the local file directory where the code
    is located. This is resolved to the original reference if installed with `-e`
    otherwise is a copy of the package.

    """
    dist = importlib.metadata.distribution(package_name)

    def normalize_package(package: str):
        return package.replace("-", "_").lower()

    # Recent versions of poetry install development packages (-e .) as direct URLs
    # https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/introduction.html
    # "Path configuration files have an extension of .pth, and each line must
    # contain a single path that will be appended to sys.path."
    package_name = normalize_package(dist.name)
    symbolic_links = [
        path
        for path in (dist.files or [])
        if path.name.lower() == f"{package_name}.pth"
    ]
    dist_links = [
        path
        for path in (dist.files or [])
        if path.name == "direct_url.json"
        and re_search(package_name + r"-[0-9-.]+\.dist-info", path.parent.name.lower())
    ]
    explicit_links = [
        path
        for path in (dist.files or [])
        if path.parent.name.lower() == package_name
        and (
            # Sanity check that the parent is the high level project directory
            # by looking for common base files
            path.name == "__init__.py"
        )
    ]

    # The user installed code as an absolute package (ie. with pip install .) instead of
    # as a reference. There's no need to sniff for the additional package path since
    # we've already found it
    if explicit_links:
        # Right now we have a file pointer to __init__.py. Go up one level
        # to the main package directory to return a directory
        explicit_link = explicit_links[0]
        return Path(str(dist.locate_file(explicit_link.parent)))

    # Raw path will capture the path to the pyproject.toml file directory,
    # not the actual package code directory
    # Find the root, then resolve the package directory
    raw_path: Path | None = None

    if symbolic_links:
        direct_url_path = symbolic_links[0]
        raw_path = Path(str(dist.locate_file(direct_url_path.read_text().strip())))
    elif dist_links:
        dist_link = dist_links[0]
        direct_metadata = json_loads(dist_link.read_text())
        package_path = "/" + direct_metadata["url"].lstrip("file://").lstrip("/")
        raw_path = Path(str(dist.locate_file(package_path)))

    if not raw_path:
        raise ValueError(
            f"Could not find a valid path for package {dist.name}, found files: {dist.files}"
        )

    # Sniff for the presence of the code directory
    for path in raw_path.iterdir():
        if path.is_dir() and normalize_package(path.name) == package_name:
            return path

    raise ValueError(
        f"No matching package found in root path: {raw_path} {list(raw_path.iterdir())}"
    )



from abc import ABC, abstractmethod
from os import environ
from pathlib import Path
from subprocess import Popen


class EnvironmentBase(ABC):
    """
    Base class for backend environments where we'll install the python
    dependencies in the remote project, so we avoid a system wide install.

    """

    def __init__(self):
        self.global_env = {
            "HOME": environ.get("HOME", ""),
            "GOROOT": environ.get("GOROOT", ""),
            "GOPATH": environ.get("GOPATH", ""),
            "GOMODCACHE": environ.get("GOMODCACHE", ""),
        }

    @abstractmethod
    def has_provider(self) -> bool:
        """
        Determines whether the environment can be executed locally
        by a given provider

        """
        pass

    @abstractmethod
    def install_project(self, project_path: Path) -> None:
        """
        Perform the installation of the python dependencies into
        the current project.

        """
        pass

    @abstractmethod
    def run_command(self, command: list[str], path: Path) -> Popen:
        """
        Run a command in the environment

        """
        pass

    @abstractmethod
    def get_env_path(self, project_path: Path) -> str:
        """
        Get the environment path. Only valid after the installation
        process has completed. Otherwise will raise an error.

        """
        pass



import json
import logging
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from io import StringIO

from mountaineer.logging import setup_logger


@contextmanager
def capture_logger_output(logger_func, *args, **kwargs):
    new_stdout, new_stderr = StringIO(), StringIO()
    with redirect_stdout(new_stdout), redirect_stderr(new_stderr):
        logger = logger_func(*args, **kwargs)
        yield logger, new_stdout, new_stderr


def test_setup_logger_name_and_level():
    logger_name = "test_logger"
    log_level = logging.DEBUG
    with capture_logger_output(setup_logger, logger_name, log_level) as (logger, _, _):
        assert logger.name == logger_name
        assert logger.level == log_level


def test_log_message_output():
    log_message = "Test message"
    logger_name = "output_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        logger.info(log_message)
        output = stdout.getvalue()
    # Since actual output includes ANSI escape codes for colors, we validate the presence of the message and level
    assert log_message in output
    assert "INFO" in output


def test_timestamp():
    logger_name = "format_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        logger.warning("Warning test")
        output = stdout.getvalue()
    log_output = json.loads(output.strip())
    assert "timestamp" in log_output


def test_exception_logging():
    logger_name = "exception_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        try:
            raise ValueError("Test exception")
        except ValueError:
            logger.exception("This is an exception")
        output = stdout.getvalue()
        log_output = json.loads(output.strip())
        assert "exception" in log_output
        assert "Test exception" in log_output["exception"]



import collections
import collections.abc
import typing
import warnings
from enum import Enum
from inspect import (
    isclass,
    ismethod,
)
from json import loads as json_loads
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    get_args,
    get_origin,
)

import starlette.responses
from fastapi.responses import JSONResponse
from inflection import camelize
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.exceptions import APIException
from mountaineer.render import FieldClassDefinition, Metadata, RenderBase, RenderNull


class FunctionActionType(Enum):
    RENDER = "render"
    SIDEEFFECT = "sideeffect"
    PASSTHROUGH = "passthrough"


class ResponseModelType(Enum):
    SINGLE_RESPONSE = "SINGLE_RESPONSE"
    ITERATOR_RESPONSE = "ITERATOR_RESPONSE"


class FunctionMetadata(BaseModel):
    function_name: str
    action_type: FunctionActionType

    # Specified for sideeffects, where all data shouldn't be update. Limits the
    # update to fields defined in this tuple.
    reload_states: tuple[
        FieldClassDefinition, ...
    ] | None | MountaineerUnsetValue = MountaineerUnsetValue()

    # Defines the data schema returned from the function that will be included in the
    # response payload sent to the client. This might be used for either passthrough
    # or sideeffect
    passthrough_model: Type[
        BaseModel
    ] | None | MountaineerUnsetValue = MountaineerUnsetValue()
    exception_models: list[
        Type[APIException]
    ] | None | MountaineerUnsetValue = MountaineerUnsetValue()
    media_type: str | None | MountaineerUnsetValue = MountaineerUnsetValue()
    is_raw_response: bool = False

    # Render type, defines the data model that is returned by the render typehint
    # If "None", the user has explicitly stated that no render model is returned
    render_model: Type[
        RenderBase
    ] | None | MountaineerUnsetValue = MountaineerUnsetValue()

    # Inserted by the render decorator
    return_model: Type[BaseModel] | MountaineerUnsetValue = MountaineerUnsetValue()

    model_config = {
        "arbitrary_types_allowed": True,
    }

    #
    # Accessors for polymorphic variables
    # These convenience methods are used to ensure client callers that this function metadata
    # does have the expected values. They should only be used in cases that you know based on runtime
    # guarantees that certain functions will have certain attributes.
    #
    def get_reload_states(self) -> tuple[FieldClassDefinition, ...] | None:
        if isinstance(self.reload_states, MountaineerUnsetValue):
            raise ValueError("Reload states not set")
        return self.reload_states

    def get_render_model(self) -> Type[RenderBase] | None:
        if isinstance(self.render_model, MountaineerUnsetValue):
            raise ValueError("Render model not set")
        return self.render_model or RenderNull

    def get_passthrough_model(self) -> Type[BaseModel] | None:
        if isinstance(self.passthrough_model, MountaineerUnsetValue):
            raise ValueError("Passthrough model not set")
        return self.passthrough_model

    def get_exception_models(self) -> list[Type[APIException]] | None:
        if isinstance(self.exception_models, MountaineerUnsetValue):
            raise ValueError("Exception models not set")
        return self.exception_models

    def get_media_type(self) -> str | None:
        if isinstance(self.media_type, MountaineerUnsetValue):
            raise ValueError("Media type not set")
        return self.media_type

    def get_is_raw_response(self) -> bool:
        return self.is_raw_response

    def get_return_model(self) -> Type[BaseModel]:
        if isinstance(self.return_model, MountaineerUnsetValue):
            raise ValueError("Return model not set")
        return self.return_model


METADATA_ATTRIBUTE = "_mountaineer_metadata"


def init_function_metadata(fn: Callable, action_type: FunctionActionType):
    if ismethod(fn):
        fn = fn.__func__  # type: ignore

    function_name = fn.__name__
    metadata = FunctionMetadata(function_name=function_name, action_type=action_type)
    setattr(fn, METADATA_ATTRIBUTE, metadata)
    return metadata


def get_function_metadata(fn: Callable) -> FunctionMetadata:
    if ismethod(fn):
        fn = fn.__func__  # type: ignore

    if not hasattr(fn, METADATA_ATTRIBUTE):
        raise AttributeError(f"Function {fn.__name__} does not have metadata")

    metadata = getattr(fn, METADATA_ATTRIBUTE)

    if not isinstance(metadata, FunctionMetadata):
        raise AttributeError(f"Function {fn.__name__} has invalid metadata")

    return metadata


def annotation_is_metadata(annotation: type | None):
    if not annotation:
        return

    return annotation == Metadata or annotation == Optional[Metadata]


def fuse_metadata_to_response_typehint(
    metadata: FunctionMetadata,
    render_model: Type[RenderBase] | None,
) -> Type[BaseModel]:
    """
    Functions can either be marked up with side effects, explicit responses, or both.
    This function merges them into the expected output payload so we can typehint the responses.
    """
    passthrough_fields = {}
    sideeffect_fields = {}

    if metadata.passthrough_model is not None and not isinstance(
        metadata.passthrough_model, MountaineerUnsetValue
    ):
        passthrough_fields = {**metadata.passthrough_model.model_fields}

    if metadata.action_type == FunctionActionType.SIDEEFFECT and render_model:
        # By default, reload all fields
        sideeffect_fields = {**render_model.model_fields}

        # Ignore the metadata since this shouldn't be passed during sideeffects
        sideeffect_fields = {
            field_name: field_definition
            for field_name, field_definition in sideeffect_fields.items()
            if not annotation_is_metadata(field_definition.annotation)
        }

        if metadata.reload_states is not None and not isinstance(
            metadata.reload_states, MountaineerUnsetValue
        ):
            # Make sure this class actually aligns to the response model
            # If not the user mis-specified the reload states
            #
            # We allow the reload state to be a subclass of the render model, in case the method
            # was originally defined in the superclass. This will result in us sending a subset
            # of the child controller's fields - but since our differential update is based on the
            # original state and modified, this will resolve correctly for the client
            reload_classes = {field.root_model for field in metadata.reload_states}
            reload_keys = {field.key for field in metadata.reload_states}
            if len(reload_classes) != 1 or not issubclass(
                render_model, next(iter(reload_classes))
            ):
                raise ValueError(
                    f"Reload states {reload_classes} do not align to response model {render_model}"
                )
            sideeffect_fields = {
                field_name: field_definition
                for field_name, field_definition in render_model.model_fields.items()
                if field_name in reload_keys
            }

    base_response_name = camelize(metadata.function_name) + "Response"
    base_response_params = {}

    if passthrough_fields:
        base_response_params["passthrough"] = (
            create_model(
                base_response_name + "Passthrough",
                **{
                    field_name: (field_definition.annotation, field_definition)  # type: ignore
                    for field_name, field_definition in passthrough_fields.items()
                },
            ),
            FieldInfo(alias="passthrough"),
        )

    if sideeffect_fields:
        base_response_params["sideeffect"] = (
            create_model(
                base_response_name + "SideEffect",
                **{
                    field_name: (field_definition.annotation, field_definition)  # type: ignore
                    for field_name, field_definition in sideeffect_fields.items()
                },
            ),
            FieldInfo(alias="sideeffect"),
        )

    model: Type[BaseModel] = create_model(
        base_response_name,
        **base_response_params,  # type: ignore
    )
    return model


def handle_explicit_responses(
    dict_payload: dict[str, Any],
):
    """
    Wrapper to allow actions to respond with an explicit JSONResponse, or a dictionary. This lets
    both sideeffects and passthrough payloads to inject header metadata that otherwise can't be captured
    in a regular BaseModel.

    Since the eventual result of an action is a combined sideeffect+passthrough payload, we need to
    merge the final payload into the explicit response.

    """
    responses = [
        (key, response)
        for key, response in dict_payload.items()
        if isinstance(response, JSONResponse)
    ]

    if len(responses) > 1:
        raise ValueError(f"Multiple conflicting responses returned: {responses}")

    if len(responses) == 0:
        return dict_payload

    response_key, response = responses[0]
    dict_payload[response_key] = json_loads(response.body)

    # Now inject the newly formatted response into the response object
    return JSONResponse(
        content=dict_payload,
        status_code=response.status_code,
        headers={
            key: value
            for key, value in response.headers.items()
            if key not in {"content-length", "content-type"}
        },
    )


def extract_response_model_from_signature(
    func: Callable, explicit_response: Type[BaseModel] | None = None
):
    typehinted_response = func.__annotations__.get("return", MountaineerUnsetValue())
    if explicit_response:
        warnings.warn(
            (
                "The response_model argument is deprecated. Instead, just typehint your function explicitly:\n"
                "def my_function() -> MyResponseModel:"
            ),
            DeprecationWarning,
            stacklevel=2,
        )

    if explicit_response:
        return explicit_response, ResponseModelType.SINGLE_RESPONSE

    if isinstance(typehinted_response, MountaineerUnsetValue):
        # This will be converted into a ValueError in the future
        # For now, backwards compatible with the old markup
        warnings.warn(
            (
                f"You must typehint the return value of your `{func}` with either a BaseModel or None.\n"
                "We will stop inferring `None` as a response model in the future."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return None, ResponseModelType.SINGLE_RESPONSE

    return extract_model_from_decorated_types(typehinted_response)


def extract_model_from_decorated_types(
    type_hint: Any,
) -> tuple[Type[BaseModel] | None, ResponseModelType]:
    """
    Support response_model typehints like Iterator[Type[BaseModel]] and AsyncIterator[Type[BaseModel]].

    """
    origin_type = get_origin(type_hint)

    if type_hint is None:
        return None, ResponseModelType.SINGLE_RESPONSE
    elif isclass(type_hint) and issubclass(type_hint, BaseModel):
        return type_hint, ResponseModelType.SINGLE_RESPONSE
    elif origin_type in (
        typing.Iterator,
        typing.AsyncIterator,
        # At runtime our types are sometimes instantiated as collections.abc objects
        collections.abc.Iterator,
        collections.abc.AsyncIterator,
    ):
        args = get_args(type_hint)
        if args and issubclass(args[0], BaseModel):
            return args[0], ResponseModelType.ITERATOR_RESPONSE
        raise ValueError(
            f"Invalid response_model typehint for iterator action: {type_hint} {origin_type} {args}"
        )
    elif isclass(type_hint) and issubclass(type_hint, starlette.responses.Response):
        # No pydantic model to include in the API schema, instead the endpoint
        # will just return the raw value
        return None, ResponseModelType.SINGLE_RESPONSE

    raise ValueError(
        f"Invalid response_model typehint for standard action: {type_hint}"
    )



import functools
from collections import OrderedDict
from hashlib import sha256
from json import dumps as json_dumps
from typing import Any, Callable

from pydantic import BaseModel

from mountaineer.logging import LOGGER


class LRUCache:
    def __init__(self, capacity: int, max_size_bytes: int | None):
        self.cache: OrderedDict[str, Any] = OrderedDict()
        self.capacity = capacity
        self.max_size_bytes = max_size_bytes

    def get(self, key: str):
        if key not in self.cache:
            return None
        self.cache.move_to_end(key)
        return self.cache[key]

    def put(self, key: str, value: Any, size_bytes: int):
        if self.max_size_bytes and size_bytes > self.max_size_bytes:
            LOGGER.warning(
                f"Skipping cache for {key} as item exceeds the max size limit."
            )
            return
        if key in self.cache:
            self.cache.move_to_end(key)
        self.cache[key] = value
        if len(self.cache) > self.capacity:
            self.cache.popitem(last=False)

    def clear(self):
        self.cache.clear()


def serialize_args(args, kwargs):
    """
    Serialize function arguments to a JSON-compatible format.
    """
    serialized: list[Any | tuple[str, Any]] = []
    for arg in args:
        if isinstance(arg, BaseModel):
            serialized.append(arg.model_dump_json())
        else:
            serialized.append(arg)
    for key, value in kwargs.items():
        if isinstance(value, BaseModel):
            serialized.append((key, value.model_dump_json()))
        else:
            serialized.append((key, value))
    return json_dumps(serialized, sort_keys=True)


def extended_lru_cache(maxsize: int, max_size_mb: float | None = None):
    """
    Main entrypoint to our custom LRU cache. Unlike the standard python version,
    this has special handling for:
    - Pydantic BaseModels, converts values to json to ensure we can hash all of the values
    - A max_size_mb parameter to limit the size of each element of the cache. If a new
        request/response set of values exceeds this size, it will not be cached.

    Will inject a `use_cache` optional argument to the function signature too, so you can
    disable caching per request if needed.

    """
    max_size_bytes = int(max_size_mb * 1024 * 1024) if max_size_mb is not None else None

    def decorator(func: Callable):
        cache = LRUCache(capacity=maxsize, max_size_bytes=max_size_bytes)

        @functools.wraps(func)
        def wrapper(*args, use_cache=True, **kwargs):
            serialized = serialize_args(args, kwargs)
            hash_key = sha256(serialized.encode()).hexdigest()

            if use_cache:
                if (result := cache.get(hash_key)) is not None:
                    return result

            result = func(*args, **kwargs)

            # Serialize result to check size
            if use_cache:
                serialized_result = json_dumps(result)
                size_bytes = len(serialized_result.encode("utf-8"))
                cache.put(hash_key, result, size_bytes)

            return result

        setattr(wrapper, "_cache", cache)
        return wrapper

    return decorator



import asyncio
from os import environ
from pathlib import Path
from subprocess import PIPE
from tempfile import TemporaryDirectory

from mountaineer.controller import ControllerBase
from mountaineer.js_compiler.base import ClientBuilderBase, ClientBundleMetadata
from mountaineer.js_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath


class PostCSSBundler(ClientBuilderBase):
    """
    Support PostCSS processing for CSS files.

    - Assumes postcss-cli is installed in the primary project's package root (ie.
      the ci_webapp/views/node_modules directory)
    - Assumes that the css file under consideration has a root postcss.config.js file
      within its own view path

    """

    async def handle_file(
        self,
        file_path: ManagedViewPath,
        controller: ControllerBase | None,
        metadata: ClientBundleMetadata,
    ):
        # If this is a CSS file we try to process it
        if file_path.suffix not in {".css", ".scss"}:
            return

        root_path = file_path.get_package_root_link()
        built_css = await self.process_css(file_path)
        (
            root_path.get_managed_static_dir(tmp_build=True)
            / self.get_style_output_name(file_path)
        ).write_text(built_css)

    async def process_css(self, css_path: ManagedViewPath) -> str:
        """
        Process a CSS file using PostCSS and output the transformed contents.
        """
        is_installed, cli_path = self.postcss_is_installed(
            css_path.get_package_root_link()
        )
        if not is_installed:
            raise EnvironmentError(
                "postcss-cli is not installed in the specified view_root_path. Install it with:\n"
                "$ npm install -D postcss postcss-cli"
            )

        with TemporaryDirectory() as temp_dir_name:
            temp_dir_path = Path(temp_dir_name)
            output_path = temp_dir_path / "output.css"

            command = [str(cli_path), str(css_path), "-o", str(output_path)]
            process = await asyncio.create_subprocess_exec(
                *command,
                stdout=PIPE,
                stderr=PIPE,
                # postcss won't find the config file if we don't set the cwd
                # we assume this is in each view's root directory
                cwd=css_path.get_root_link(),
                env={
                    **environ,
                    # We expect the main package root will be the one with node modules installed
                    "NODE_PATH": str(css_path.get_package_root_link() / "node_modules"),
                },
            )

            stdout, stderr = await process.communicate()

            if stdout.strip():
                LOGGER.info(stdout.decode())
            if stderr.strip():
                LOGGER.warning(stderr.decode())

            if process.returncode != 0:
                raise BuildProcessException(f"postcss error: {stderr.decode()}")

            return output_path.read_text()

    def get_style_output_name(self, original_stylesheet_path: ManagedViewPath) -> str:
        """
        Given a path to an original stylesheet, return the name of the compiled css file

        original_stylesheet_path: "path/to/styles.scss"
        output: "path_to_styles.css"
        """
        root_path = original_stylesheet_path.get_root_link()
        relative_path = original_stylesheet_path.relative_to(root_path)

        # Convert the parent portions of the path to make sure our final output
        # filename is unique
        unique_path = "_".join(
            [*relative_path.parent.parts, original_stylesheet_path.name]
        )
        return str(Path(unique_path).with_suffix(".css"))

    def postcss_is_installed(self, view_root_path: Path):
        # Adjust the check to look for local installation of postcss-cli, which we currently
        # require at the CLI bridge
        expected_path = view_root_path / "node_modules" / ".bin" / "postcss"
        return (
            expected_path.exists() and expected_path.is_file(),
            expected_path,
        )



from types import ModuleType
from typing import overload

from click import secho
from fastapi import Depends
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncEngine

from mountaineer.database import DatabaseDependencies, SQLModel
from mountaineer.dependencies import get_function_dependencies


@overload
async def handle_createdb(model_module: ModuleType) -> None:
    ...


@overload
async def handle_createdb(models: list[SQLModel]) -> None:
    ...


async def handle_createdb(*args, **kwargs):
    """
    Strictly speaking, passing a list of models isn't required for this function. We'll happily
    accept being called with `handle_createdb()`. We just encourage an explicit passing of either
    the models module or the SQLModels themselves to make sure they are in-scope of the table
    registry when this function is run. This is how we determine which tables to create at runtime.

    :param model_module: The module containing the SQLModels to create

    :param models: An explicit list of SQLModels to create

    """

    async def run_migrations(
        engine: AsyncEngine = Depends(DatabaseDependencies.get_db),
    ):
        async with engine.begin() as connection:
            await connection.run_sync(SQLModel.metadata.create_all)

            # Log the tables that were created
            result = await connection.execute(
                text(
                    """
                SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = 'public'
                """
                )
            )
            tables = "\n".join([f"* {table[0]}" for table in result.fetchall()])
            secho(f"Database tables created:\n{tables}", fg="green")

    async with get_function_dependencies(callable=run_migrations) as values:
        await run_migrations(**values)



from dataclasses import dataclass
from pathlib import Path
from re import finditer as re_finditer, sub
from time import monotonic_ns

from pydantic import BaseModel

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.logging import LOGGER


@dataclass
class ValueMask:
    mask: int
    right_padding: int


class SourceMapSchema(BaseModel):
    version: int
    sources: list[str]
    names: list[str]
    mappings: str
    sourcesContent: list[str] | None = None
    sourceRoot: str | None = None
    file: str | None = None


class SourceMapParser:
    """
    Parse sourcemaps according to the official specification:
    https://sourcemaps.info/spec.html

    """

    def __init__(self, path: str | Path):
        """
        :param relative_to: If specified, will output source paths relative
        to this path.

        """
        self.path = Path(path)

        self.source_map: SourceMapSchema | None = None

        # { (line, column) : MapMetadata }
        self.parsed_mappings: dict[
            tuple[int, int], mountaineer_rs.MapMetadata
        ] | None = None

    def parse(self):
        """
        Parse the source map file and build up the internal mappings. This is
        deterministic with respect to the initialized source map path, so this
        will be a no-op if it's already been run.

        """
        # If we've already parsed this file, don't do it again
        if self.parsed_mappings is not None:
            return

        start_parse = monotonic_ns()
        self.source_map = SourceMapSchema.model_validate_json(
            Path(self.path).read_text()
        )
        LOGGER.debug(f"Parsed source map in {(monotonic_ns() - start_parse)/1e9:.2f}s")

        start_parse = monotonic_ns()
        self.parsed_mappings = mountaineer_rs.parse_source_map_mappings(
            self.source_map.mappings
        )
        LOGGER.debug(f"Parsed mappings in {(monotonic_ns() - start_parse)/1e9:.2f}s")

    def get_original_location(self, line: int, column: int):
        """
        For a compiled line and column, return the original line and column where they appeared
        in the pre-built file.

        :param line: The line number in the compiled file
        :param column: The column number in the compiled file

        """
        if self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        return self.parsed_mappings.get((line, column))

    def map_exception(self, exception: str) -> str:
        """
        Given a JS stack exception, try to map it to the original files and line numbers

        :param exception: The exception string to map

        :return: The exception string with the original file and line numbers. Note that some
            exception stack traces may not be mappable, and will be left as-is.

        """
        if self.source_map is None or self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        # Build up the replacements all at once, since the matched indexes will be tied
        # to the original string
        text_replacements: dict[tuple[int, int], str] = {}
        for match in re_finditer(r"\(([<>A-Za-z0-9]+?):(\d+?):(\d+?)\)", exception):
            original_match = self.get_original_location(
                int(match.group(2)), int(match.group(3))
            )

            if original_match and original_match.source_index is not None:
                text_replacements[match.span(1)] = self.convert_relative_path(
                    self.source_map.sources[original_match.source_index]
                )
                text_replacements[match.span(2)] = str(original_match.source_line)
                text_replacements[match.span(3)] = str(original_match.source_column)

        # Sort in reverse order based on their start index to ensure that modifying parts of the string
        # doesn't affect the positions of parts that haven't been modified yet
        sorted_replacements = sorted(
            text_replacements.items(), key=lambda x: x[0][0], reverse=True
        )

        for (start, end), replacement in sorted_replacements:
            exception = exception[:start] + replacement + exception[end:]

        return exception

    def convert_relative_path(self, absolute_path: str):
        """
        Absolute paths are convenient for internal use since they fully qualify
        a given file. However, for display they often get long and repetitive across
        multiple lines. This function will convert an absolute path to a relative path
        if it's within the same directory as the current working directory.

        :param absolute_path: The absolute path to convert

        :return: The relative path if it's within the current working directory, otherwise
            the unmodified absolute path.

        """
        source_path = Path(absolute_path)

        if source_path.is_relative_to(Path.cwd()):
            return "./" + str(source_path.relative_to(Path.cwd()))

        return absolute_path


def get_cleaned_js_contents(contents: str):
    """
    Strip all single or multiline comments, since these can be dynamically generated
    metadata and can change without the underlying logic changing.
    """
    return mountaineer_rs.strip_js_comments(contents).strip()


def update_source_map_path(contents: str, new_path: str):
    """
    Updates the source map path to the new path, since the path is dynamic.
    """
    return sub(r"sourceMappingURL=(.*?).map", f"sourceMappingURL={new_path}", contents)



import multiprocessing
from dataclasses import dataclass
from pathlib import Path
from threading import Thread
from time import monotonic_ns
from typing import Any

from inflection import underscore

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.controller import ControllerBase
from mountaineer.js_compiler.base import ClientBuilderBase, ClientBundleMetadata
from mountaineer.js_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, generate_relative_import


@dataclass
class JSBundle:
    temp_path: Path
    client_entrypoint_path: Path
    server_entrypoint_path: Path

    # Also pass along the raw parameters that we were called with
    file_path: ManagedViewPath
    controller: ControllerBase
    metadata: ClientBundleMetadata


@dataclass
class BundleOutput:
    client_entrypoint_path: Path
    client_compiled_contents: str
    client_source_map_contents: str
    server_entrypoint_path: Path
    server_compiled_contents: str
    server_source_map_contents: str


@dataclass
class CompiledOutput:
    success: bool
    exception_type: str | None = None
    exception_message: str | None = None


class JavascriptBundler(ClientBuilderBase):
    """
    Compile the client-written tsx/jsx to raw javascript files for execution as part
    of the SSR pipeline and client hydration.
    """

    def __init__(
        self,
        root_element: str = "root",
        environment: str = "development",
        tmp_dir: Path | None = None,
    ):
        super().__init__(tmp_dir=tmp_dir)

        self.root_element = root_element
        self.environment = environment

        # Pending paths to be processed
        # (client_path, server_path)
        self.pending_files: list[JSBundle] = []

    async def init_state(self, global_state):
        self.global_state = global_state

        # Launch a thread that will handle our build queue, potentially across
        # multiple subprocesses
        build_input_queue: multiprocessing.Queue[tuple[Any]] = multiprocessing.Queue()
        build_output_queue: multiprocessing.Queue[
            CompiledOutput
        ] = multiprocessing.Queue()

        global_state["js_bundler_input"] = build_input_queue
        global_state["js_bundler_output"] = build_output_queue

        thread = Thread(
            target=self.process_pending_files,
            args=(build_input_queue, build_output_queue),
            daemon=True,
        )
        thread.start()

    def process_pending_files(
        self, input_queue: multiprocessing.Queue, output_queue: multiprocessing.Queue
    ):
        LOGGER.debug("Spawning process_pending_files")
        while True:
            try:
                payload = input_queue.get()
            except EOFError:
                LOGGER.debug("Exiting process_pending_files")
                return

            LOGGER.debug(f"Got JS build payload: {len(payload)}")
            if payload is None:
                break

            build_params = [
                mountaineer_rs.BuildContextParams(*params) for params in payload
            ]

            start = monotonic_ns()
            try:
                # TODO: Right now this raises pyo3_runtime.PanicException, which isn't caught
                # appropriately. Our try/except block should be catching this.
                mountaineer_rs.build_javascript(build_params)
            except Exception as e:
                LOGGER.error(f"Error building JS: {e}")
                output_queue.put(
                    CompiledOutput(
                        success=False,
                        exception_type=e.__class__.__name__,
                        exception_message=str(e),
                    )
                )
                continue

            LOGGER.debug(
                f"Processed payload in {(monotonic_ns() - start) / 1e9} seconds"
            )

            output_queue.put(CompiledOutput(success=True))

    async def start_build(self):
        self.pending_files = []

        # Make sure we have been initialized
        if self.global_state is None:
            await self.init_state({})

    async def handle_file(
        self,
        file_path: ManagedViewPath,
        controller: ControllerBase | None,
        metadata: ClientBundleMetadata,
    ):
        if controller is None:
            # Require a controller for our bundling
            return None
        if file_path.suffix not in [".tsx", ".jsx"]:
            # We only know how to parse tsx and jsx files
            return None

        # Now we can process the files in bulk
        payload = self.generate_js_bundle(
            file_path=file_path, controller=controller, metadata=metadata
        )
        self.pending_files.append(payload)

    async def finish_build(self):
        if not self.global_state:
            raise ValueError("JavascriptBundler has not been initialized")

        build_params = []
        for payload in self.pending_files:
            controller_base = underscore(payload.controller.__class__.__name__)

            root_path = payload.file_path.get_package_root_link()
            static_dir = root_path.get_managed_static_dir(tmp_build=True)
            ssr_dir = root_path.get_managed_ssr_dir(tmp_build=True)

            # Client entrypoint config
            # All these tuple arguments map to the input __init__ arguments
            # for mountaineer_rs.BuildContextParams
            build_params.append(
                (
                    str(payload.client_entrypoint_path),
                    str(payload.temp_path / "node_modules"),
                    self.environment,
                    payload.metadata.live_reload_port
                    if payload.metadata.live_reload_port
                    else 0,
                    False,
                    controller_base,
                    str(static_dir),
                )
            )
            # Server entrypoint config
            build_params.append(
                (
                    str(payload.server_entrypoint_path),
                    str(payload.temp_path / "node_modules"),
                    self.environment,
                    0,
                    True,
                    controller_base,
                    str(ssr_dir),
                )
            )

        # Send the build params to the queue
        self.global_state["js_bundler_input"].put(build_params)
        output_payload = self.global_state["js_bundler_output"].get()

        if not output_payload.success:
            if output_payload.exception_type == "ValueError":
                raise BuildProcessException(output_payload.exception_message)
            raise ValueError(
                f"Error building JS, unknown error: {output_payload.exception_message}"
            )

    def generate_js_bundle(
        self,
        file_path: ManagedViewPath,
        controller: ControllerBase,
        metadata: ClientBundleMetadata,
    ) -> JSBundle:
        # Before starting, make sure all the files are valid
        self.validate_page(
            page_path=file_path, view_root_path=file_path.get_root_link()
        )

        # Since this directory should persist between build lifecycles, we both use a unchanging
        # directory name and one that will be unique across different controllers
        temp_dir_path = self.tmp_dir / controller.__class__.__name__
        temp_dir_path.mkdir(exist_ok=True)

        layout_paths = self.sniff_for_layouts(
            page_path=file_path, view_root_path=file_path.get_root_link()
        )

        # The same endpoint definition is used for both SSR and the client build
        synthetic_payload = self.build_synthetic_endpoint(
            page_path=file_path,
            layout_paths=layout_paths,
            # This output path should be relative to the `synthetic_client` and `synthetic_server`
            # entrypoint files
            output_path=temp_dir_path,
        )

        client_entrypoint = self.build_synthetic_client_page(*synthetic_payload)
        ssr_entrypoint = self.build_synthetic_ssr_page(*synthetic_payload)

        client_entrypoint_path = temp_dir_path / "synthetic_client.tsx"
        server_entrypoint_path = temp_dir_path / "synthetic_server.tsx"

        self.link_project_files(
            view_root_path=file_path.get_package_root_link(),
            temp_dir_path=temp_dir_path,
        )
        client_entrypoint_path.write_text(client_entrypoint)
        server_entrypoint_path.write_text(ssr_entrypoint)

        return JSBundle(
            temp_path=temp_dir_path,
            client_entrypoint_path=client_entrypoint_path,
            server_entrypoint_path=server_entrypoint_path,
            file_path=file_path,
            controller=controller,
            metadata=metadata,
        )

    def build_synthetic_client_page(
        self,
        synthetic_imports: list[str],
        synthetic_endpoint: str,
        synthetic_endpoint_name: str,
    ):
        lines: list[str] = []

        # Assume the client page is always being called from a page that has been
        # initially rendered by SSR
        lines.append("import * as React from 'react';")
        lines.append("import { hydrateRoot } from 'react-dom/client';")
        lines += synthetic_imports

        lines.append(synthetic_endpoint)

        lines.append(
            f"const container = document.getElementById('{self.root_element}');"
        )
        lines.append(f"hydrateRoot(container, <{synthetic_endpoint_name} />);")

        return "\n".join(lines)

    def build_synthetic_ssr_page(
        self,
        synthetic_imports: list[str],
        synthetic_endpoint: str,
        synthetic_endpoint_name: str,
    ):
        lines: list[str] = []

        lines.append("import * as React from 'react';")
        lines.append("import { renderToString } from 'react-dom/server';")
        lines += synthetic_imports

        lines.append(synthetic_endpoint)

        lines.append(
            f"export const Index = () => renderToString(<{synthetic_endpoint_name} />);"
        )

        return "\n".join(lines)

    def build_synthetic_endpoint(
        self, *, page_path: ManagedViewPath, layout_paths: list[Path], output_path: Path
    ):
        """
        Following the Next.js syntax, layouts wrap individual pages in a top-down order. Here we
        create a synthetic page that wraps the actual page in the correct order.
        The output is a valid React file that acts as the page entrypoint
        for the `rootElement` ID in the DOM.

        """
        # All import paths have to be a relative path from the scratch directory
        # to the original file
        import_paths: list[str] = []

        static_api_path = (
            page_path.get_package_root_link().get_managed_code_dir() / "live_reload.ts"
        )
        import_paths.append(
            f"import mountLiveReload from '{generate_relative_import(output_path, static_api_path)}';"
        )

        import_paths.append(
            f"import Page from '{generate_relative_import(output_path, page_path)}';"
        )

        for i, layout_path in enumerate(layout_paths):
            import_paths.append(
                f"import Layout{i} from '{generate_relative_import(output_path, layout_path)}';"
            )

        # The synthetic endpoint is a function that returns a React component
        entrypoint_name = "Entrypoint"
        content_lines = [
            f"const {entrypoint_name} = () => {{",
            # This hook will only run in the browser, in dev mode
            # Otherwise it will short-circuit so it won't apply to production
            "mountLiveReload({});" "return (",
            *[f"<Layout{i}>" for i in range(len(layout_paths))],
            "<Page />",
            *[f"</Layout{i}>" for i in reversed(range(len(layout_paths)))],
            ");",
            "};",
        ]

        return import_paths, "\n".join(content_lines), entrypoint_name

    def link_project_files(self, *, view_root_path: Path, temp_dir_path: Path):
        """
        Javascript packages define a variety of build metadata in the root directory
        of the project (tsconfig.json, package.json, etc). Since we're running our esbuild pipeline
        in a temporary directory, we need to copy over the key files. We use a symbolic link
        to avoid copying the files over.
        """
        required_to_link = ["package.json", "node_modules"]
        optional_to_link = ["tsconfig.json"]

        for file_name in required_to_link + optional_to_link:
            # Only throw an error if the file is required to exist
            if not (view_root_path / file_name).exists():
                if file_name in required_to_link:
                    raise ValueError(
                        f"Error linking, expected {file_name} to exist in {view_root_path}"
                    )
                continue

            # If the file exists, we assume it's to the correct path
            LOGGER.debug(f"Linking {file_name} to {temp_dir_path}")
            (temp_dir_path / file_name).unlink(missing_ok=True)
            (temp_dir_path / file_name).symlink_to(view_root_path / file_name)

    def sniff_for_layouts(self, *, page_path: Path, view_root_path: Path):
        """
        Given a page.tsx path, find all the layouts that apply to it.
        Returns the layout paths that are found. Orders them from the top->down
        as they expect to be rendered.

        """
        # It's easier to handle absolute paths when doing direct string comparisons
        # of the file hierarchy.
        page_path = page_path.resolve().absolute()
        view_root_path = view_root_path.resolve().absolute()

        # Starting at the page path, walk up the directory tree and yield each layout
        # that is found.
        layouts: list[Path] = []
        current_path = page_path.parent

        while current_path != view_root_path:
            layout_path_tsx = current_path / "layout.tsx"
            layout_path_jsx = current_path / "layout.jsx"

            # We shouldn't have both in the same directory
            if layout_path_tsx.exists() and layout_path_jsx.exists():
                raise ValueError(
                    f"Duplicate layout definitions: {layout_path_tsx}, {layout_path_jsx}"
                )

            if layout_path_tsx.exists():
                layouts.append(layout_path_tsx)
            if layout_path_jsx.exists():
                layouts.append(layout_path_jsx)

            current_path = current_path.parent

        # Return the layouts in the order they should be rendered
        return list(reversed(layouts))

    def validate_page(self, *, page_path: Path, view_root_path: Path):
        # Validate that we're actually calling on a path file
        if page_path.name not in {"page.tsx", "page.jsx"}:
            raise ValueError(
                f"Invalid page path; view need to be specified in a `page.tsx` file: {page_path}"
            )

        # Validate that the page_path is within the view root. The following
        # logic assumes a hierarchical relationship between the two.
        if not page_path.is_relative_to(view_root_path):
            raise ValueError(
                f"Invalid page path, not relative to view root: {page_path} (root: {view_root_path})"
            )



import logging
from contextlib import contextmanager
from json import dumps as json_dumps
from logging import Formatter, StreamHandler, getLogger
from time import monotonic_ns

from click import secho


class JsonFormatter(Formatter):
    def format(self, record):
        log_record = {
            "level": record.levelname,
            "name": record.name,
            "timestamp": self.formatTime(record, self.datefmt),
            "message": record.getMessage(),
        }
        if record.exc_info:
            log_record["exception"] = self.formatException(record.exc_info)
        return json_dumps(log_record)


class ColorHandler(StreamHandler):
    def emit(self, record):
        try:
            msg = self.format(record)
            if record.levelno == logging.WARNING:
                secho(msg, fg="yellow")
            elif record.levelno >= logging.ERROR:
                secho(msg, fg="red")
            else:
                secho(msg)
        except Exception:
            self.handleError(record)


def setup_logger(name, log_level=logging.DEBUG):
    """
    Constructor for the main logger used by Mountaineer. Provided
    convenient defaults for log level and formatting, alongside coloring
    of stdout/stderr messages and JSON fields for structured parsing.

    """
    # TODO - env driven logging configuration

    logger = getLogger(name)
    logger.setLevel(log_level)

    # Create a handler that writes log records to the standard error
    handler = ColorHandler()
    handler.setLevel(log_level)

    formatter = JsonFormatter()
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


@contextmanager
def log_time_duration(message: str):
    """
    Context manager to time a code block at runtime.

    ```python
    with log_time_duration("Long computation"):
        # Simulate work
        sleep(10)
    ```

    """
    start = monotonic_ns()
    yield
    LOGGER.debug(f"{message} : Took {(monotonic_ns() - start)/1e9:.2f}s")


LOGGER = setup_logger(__name__)



import asyncio
from collections import defaultdict
from dataclasses import asdict, dataclass
from json import dumps as json_dumps
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from time import monotonic_ns
from typing import Any

from click import secho
from fastapi import APIRouter
from inflection import camelize

from mountaineer.actions import get_function_metadata
from mountaineer.actions.fields import FunctionActionType
from mountaineer.app import AppController, ControllerDefinition
from mountaineer.client_builder.build_actions import (
    OpenAPIToTypescriptActionConverter,
)
from mountaineer.client_builder.build_links import OpenAPIToTypescriptLinkConverter
from mountaineer.client_builder.build_schemas import OpenAPIToTypescriptSchemaConverter
from mountaineer.client_builder.openapi import OpenAPIDefinition, OpenAPISchema
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)
from mountaineer.controller import ControllerBase
from mountaineer.io import gather_with_concurrency
from mountaineer.js_compiler.base import ClientBundleMetadata
from mountaineer.js_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, generate_relative_import
from mountaineer.static import get_static_path


@dataclass
class RenderSpec:
    url: str
    view_path: str
    spec: dict[Any, Any] | None


class ClientBuilder:
    """
    Main entrypoint for building the auto-generated typescript code.

    """

    def __init__(
        self,
        app: AppController,
        live_reload_port: int | None = None,
        build_cache: Path | None = None,
    ):
        self.openapi_schema_converter = OpenAPIToTypescriptSchemaConverter(
            export_interface=True
        )
        self.openapi_action_converter = OpenAPIToTypescriptActionConverter()
        self.openapi_link_converter = OpenAPIToTypescriptLinkConverter()

        self.app = app
        self.view_root = ManagedViewPath.from_view_root(app.view_root)
        self.live_reload_port = live_reload_port
        self.build_cache = build_cache

    def build(self):
        asyncio.run(self.async_build())

    async def async_build(self):
        # Avoid rebuilding if we don't need to
        if self.cache_is_outdated():
            secho("Building useServer, cache outdated...", fg="green")

            # Make sure our application definitions are in a valid state before we start
            # to build the client code
            self.validate_unique_paths()

            # Static files that don't depend on client code
            self.generate_static_files()

            # The order of these generators don't particularly matter since most TSX linters
            # won't refresh until they're all complete. However, this ordering better aligns
            # with semantic dependencies so we keep the linearity where possible.
            self.generate_model_definitions()
            self.generate_action_definitions()
            self.generate_link_shortcuts()
            self.generate_link_aggregator()
            self.generate_view_servers()
            self.generate_index_file()
        else:
            secho("useServer up to date", fg="green")

        await self.build_javascript_chunks()

        # Update the cached paths attached to the app
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            controller.resolve_paths(self.view_root, force=True)

    def generate_static_files(self):
        """
        Copy over the static files that are required for the client.

        """
        for static_filename in ["api.ts", "live_reload.ts"]:
            managed_code_dir = self.view_root.get_managed_code_dir()
            api_content = get_static_path(static_filename).read_text()
            (managed_code_dir / static_filename).write_text(api_content)

    def generate_model_definitions(self):
        """
        Generate the interface type definitions for the models. These most closely
        apply to the controller that they're defined within, so we create the files
        directly within the controller's view directory.

        """
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller

            action_spec_openapi = self.openapi_action_specs[controller]
            action_base = OpenAPIDefinition(**action_spec_openapi)

            render_spec_openapi = self.openapi_render_specs[controller]
            render_base = (
                OpenAPISchema(**render_spec_openapi.spec)
                if render_spec_openapi.spec
                else None
            )

            schemas: dict[str, str] = {}

            # Convert the render model
            if render_base:
                for (
                    schema_name,
                    component,
                ) in self.openapi_schema_converter.convert_schema_to_typescript(
                    render_base,
                    # Render models are sent server -> client, so we know they'll provide all their
                    # values in the initial payload
                    all_fields_required=True,
                ).items():
                    schemas[schema_name] = component

            # Convert all the other models defined in sideeffect routes
            for schema_name, component in action_base.components.schemas.items():
                schemas[
                    schema_name
                ] = self.openapi_schema_converter.convert_schema_to_interface(
                    component,
                    base=action_base,
                    # Don't require client data uploads to have all the fields
                    # if there's a value specified server side. Note that this will apply
                    # to both requests/response models right now, so clients might need
                    # to do additional validation on the response side to confirm that
                    # the server did send down the default values.
                    # This is in contrast to the render() where we always know the response
                    # payloads should be force required.
                    defaults_are_required=False,
                    all_fields_required=False,
                )

            # We put in one big models.ts file to enable potentially cyclical dependencies
            managed_code_dir = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()
            (managed_code_dir / "models.ts").write_text(
                "\n\n".join(
                    [
                        schema
                        for _, schema in sorted(schemas.items(), key=lambda x: x[0])
                    ]
                )
            )

    def generate_action_definitions(self):
        """
        Generate the actions for each controller. This should correspond the actions that are accessible
        via the OpenAPI schema and the internal router.

        """
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            controller_code_dir = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()
            root_code_dir = self.view_root.get_managed_code_dir()

            controller_action_path = controller_code_dir / "actions.ts"
            root_common_handler = root_code_dir / "api.ts"
            root_api_import_path = generate_relative_import(
                controller_action_path, root_common_handler
            )

            openapi_raw = self.openapi_from_controller(controller_definition)
            output_schemas, required_types = self.openapi_action_converter.convert(
                openapi_raw
            )

            chunks: list[str] = []

            chunks.append(
                f"import {{ __request, FetchErrorBase }} from '{root_api_import_path}';\n"
                + f"import type {{ {', '.join(required_types)} }} from './models';"
            )

            chunks += output_schemas.values()

            controller_action_path.write_text("\n\n".join(chunks))

    def generate_link_shortcuts(self):
        """
        Generate the local link formatters that are tied to each controller.

        """
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            controller_code_dir = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()
            root_code_dir = self.view_root.get_managed_code_dir()

            controller_links_path = controller_code_dir / "links.ts"

            root_common_handler = root_code_dir / "api.ts"
            root_api_import_path = generate_relative_import(
                controller_links_path, root_common_handler
            )
            render_route = controller_definition.render_router
            render_openapi = self.app.generate_openapi(
                routes=render_route.routes,
            )

            content = ""
            content += f"import {{ __getLink }} from '{root_api_import_path}';\n"
            content += self.openapi_link_converter.convert(render_openapi)

            controller_links_path.write_text(content)

    def generate_link_aggregator(self):
        """
        We need a global function that references each controller's link generator,
        so we can do controller->global->controller.

        """
        global_code_dir = self.view_root.get_managed_code_dir()

        import_paths: list[str] = []
        global_setters: dict[str, str] = {}

        # For each controller, import the links and export them
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            controller_code_dir = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()

            relative_import = generate_relative_import(
                global_code_dir / "links.ts",
                controller_code_dir / "links.ts",
            )

            # Avoid global namespace collisions
            local_link_function_name = (
                f"{camelize(controller.__class__.__name__)}GetLinks"
            )

            import_paths.append(
                f"import {{ getLink as {local_link_function_name} }} from '{relative_import}';"
            )
            global_setters[
                TSLiteral(camelize(controller.__class__.__name__, False))
            ] = TSLiteral(local_link_function_name)

        lines = [
            *import_paths,
            f"const linkGenerator = {python_payload_to_typescript(global_setters)};\n",
            "export default linkGenerator;",
        ]

        (global_code_dir / "links.ts").write_text("\n".join(lines))

    def generate_view_servers(self):
        """
        Generate the useServer() hooks within each local view. These will reference the main
        server provider and allow each view to access the particular server state that
        is linked to that controller.

        """
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller

            chunks: list[str] = []

            # Step 1: Interface to optionally override the current controller state
            # We want to have an inline reference to a model which is compatible with the base render model alongside
            # all sideeffect sub-models. Since we're re-declaring this in the server file, we also
            # have to bring with us all of the other sub-model imports.
            render_model_name = self.get_render_local_state(controller)

            # Step 2: Find the actions that are relevant
            controller_action_metadata = [
                metadata for _, _, metadata in controller._get_client_functions()
            ]

            # Step 2: Setup imports from the single global provider
            controller_model_path = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()
            global_server_path = self.view_root.get_managed_code_dir()
            relative_server_path = generate_relative_import(
                controller_model_path, global_server_path
            )

            chunks.append(
                "import React, { useState } from 'react';\n"
                + f"import {{ applySideEffect }} from '{relative_server_path}/api';\n"
                + f"import LinkGenerator from '{relative_server_path}/links';\n"
                + f"import {{ {render_model_name} }} from './models';\n"
                + (
                    f"import {{ {', '.join([metadata.function_name for metadata in controller_action_metadata])} }} from './actions';"
                    if controller_action_metadata
                    else ""
                )
            )

            # Step 3: Add the optional model definition - this allows any controller that returns a partial
            # side-effect to update the full model with the same typehint
            optional_model_name = f"{render_model_name}Optional"
            chunks.append(
                f"export type {optional_model_name} = Partial<{render_model_name}>;"
            )

            # Step 4: We expect another script has already injected this global `SERVER_DATA` constant. We
            # add the typehinting here just so that the IDE can be happy.
            chunks.append("declare global {\n" "var SERVER_DATA: any;\n" "}\n")

            # Step 5: Typehint the return type of the server state in case client callers
            # want to pass this to sub-functions
            chunks.append(
                f"export interface ServerState extends {render_model_name} {{\n"
                + "linkGenerator: typeof LinkGenerator;\n"
                + (
                    "\n".join(
                        [
                            f"{metadata.function_name}: typeof {metadata.function_name};"
                            for metadata in controller_action_metadata
                        ]
                    )
                    if controller_action_metadata
                    else ""
                )
                + "}\n"
            )

            # Step 6: Final implementation of the useServer() hook, which returns a subview of the overall
            # server state that's only relevant to this controller
            chunks.append(
                "export const useServer = () : ServerState => {\n"
                + f"const [ serverState, setServerState ] = useState(SERVER_DATA as {render_model_name});\n"
                # Local function to just override the current controller
                # We make sure to wait for the previous state to be set, in case of a
                # differential update
                + f"const setControllerState = (payload: {optional_model_name}) => {{\n"
                + "setServerState((state) => ({\n"
                + "...state,\n"
                + "...payload,\n"
                + "}));\n"
                + "};\n"
                + "return {\n"
                + "...serverState,\n"
                + "linkGenerator: LinkGenerator,\n"
                + ",\n".join(
                    [
                        (
                            f"{metadata.function_name}: applySideEffect({metadata.function_name}, setControllerState)"
                            if metadata.action_type == FunctionActionType.SIDEEFFECT
                            else f"{metadata.function_name}: {metadata.function_name}"
                        )
                        for metadata in controller_action_metadata
                    ]
                )
                + "}\n"
                + "};"
            )

            (controller_model_path / "useServer.ts").write_text("\n\n".join(chunks))

    def generate_index_file(self):
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            controller_code_dir = self.view_root.get_controller_view_path(
                controller
            ).get_managed_code_dir()

            chunks: list[str] = []

            chunks.append("export * from './actions';")
            chunks.append("export * from './links';")
            chunks.append("export * from './models';")
            chunks.append("export * from './useServer';")

            (controller_code_dir / "index.ts").write_text("\n".join(chunks))

    async def build_javascript_chunks(self, max_concurrency: int = 25):
        """
        Build the final javascript chunks that will render the react documents. Each page will get
        one chunk associated with it. We suffix these files with the current md5 hash of the contents to
        allow clients to aggressively cache these contents but invalidate the cache whenever the script
        contents have rebuilt in the background.

        """
        metadata = ClientBundleMetadata(
            live_reload_port=self.live_reload_port,
        )

        # Each build command is completely independent and there's some overhead with spawning
        # each process. Make use of multi-core machines and spawn each process in its own
        # management thread so we complete the build process in parallel.
        controller_tasks = [
            builder.handle_file(
                self.view_root.get_controller_view_path(
                    controller_definition.controller
                ),
                controller=controller_definition.controller,
                metadata=metadata,
            )
            for controller_definition in self.app.controllers
            for builder in self.app.builders
        ]

        # Optionally build static files the main views and plugin views
        # This allows plugins to have custom handling for different file types
        file_tasks = [
            builder.handle_file(
                path,
                controller=None,
                metadata=metadata,
            )
            for path in self.get_static_files()
            for builder in self.app.builders
        ]

        start = monotonic_ns()
        await gather_with_concurrency(
            [builder.start_build() for builder in self.app.builders],
            n=max_concurrency,
        )
        LOGGER.debug(f"Builder launch took {(monotonic_ns() - start) / 1e9}s")

        start = monotonic_ns()
        results = await gather_with_concurrency(
            controller_tasks + file_tasks, n=max_concurrency, catch_exceptions=True
        )
        LOGGER.debug(f"Builder tasks took {(monotonic_ns() - start) / 1e9}s")

        start = monotonic_ns()
        await gather_with_concurrency(
            [builder.finish_build() for builder in self.app.builders], n=max_concurrency
        )
        LOGGER.debug(f"Builder finish took in {(monotonic_ns() - start) / 1e9}s")

        # Go through the exceptions, logging the build errors explicitly
        has_build_error = False
        final_exception: str = ""
        for result in results:
            if isinstance(result, Exception):
                has_build_error = True
                if isinstance(result, BuildProcessException):
                    secho(f"Build error: {result}", fg="red")
                final_exception += str(result)

        if has_build_error:
            raise BuildProcessException(final_exception)

        self.move_build_artifacts_into_project()

    def move_build_artifacts_into_project(self):
        """
        Now that we build has completed, we can clear out the old files and replace it
        with the thus-far temporary files

        This cleans up old controllers in the case that they were deleted, and prevents
        outdated md5 content hashes from being served

        """
        with TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)
            tmp_static_dir = tmp_path / "static"
            tmp_ssr_dir = tmp_path / "ssr"

            # Since the tmp builds are within their parent folder, we need to move
            # them out of the way before we clear
            self.view_root.get_managed_static_dir(tmp_build=True).rename(tmp_static_dir)
            self.view_root.get_managed_ssr_dir(tmp_build=True).rename(tmp_ssr_dir)

            static_dir = self.view_root.get_managed_static_dir()
            ssr_dir = self.view_root.get_managed_ssr_dir()
            for clear_dir in [static_dir, ssr_dir]:
                if clear_dir.exists():
                    rmtree(clear_dir)
                clear_dir.mkdir(parents=True)

            # Final move
            tmp_static_dir.rename(self.view_root.get_managed_static_dir())
            tmp_ssr_dir.rename(self.view_root.get_managed_ssr_dir())

    def cache_is_outdated(self):
        """
        Determines if our last build is outdated and we need to rebuild the client. Running
        this function will also update the cache to the current state.

        """
        # We need to rebuild every time
        if not self.build_cache:
            return True

        cached_metadata = self.build_cache / "client_builder_openapi.json"

        cached_contents = {
            controller_definition.controller.__class__.__name__: {
                "action": self.openapi_action_specs[controller_definition.controller],
                "render": asdict(
                    self.openapi_render_specs[controller_definition.controller]
                ),
            }
            for controller_definition in self.app.controllers
        }
        cached_str = json_dumps(cached_contents, sort_keys=True)

        if not cached_metadata.exists():
            cached_metadata.write_text(cached_str)
            return True

        if cached_metadata.read_text() != cached_str:
            cached_metadata.write_text(cached_str)
            return True

        return False

    def get_static_files(self):
        ignore_directories = ["_ssr", "_static", "_server", "node_modules"]

        for view_root in self.get_all_root_views():
            for dir_path, _, filenames in view_root.walk():
                for filename in filenames:
                    if any(
                        [
                            directory in dir_path.parts
                            for directory in ignore_directories
                        ]
                    ):
                        continue
                    yield dir_path / filename

    def validate_unique_paths(self):
        """
        Validate that all controller paths are unique. Otherwise we risk stomping
        on other server metadata that has already been written.

        """
        # Validation 1: Ensure that all view paths are unique
        view_counts = defaultdict(list)
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            view_counts[
                self.view_root.get_controller_view_path(controller).parent
            ].append(controller)
        duplicate_views = [
            (view, controllers)
            for view, controllers in view_counts.items()
            if len(controllers) > 1
        ]

        if duplicate_views:
            raise ValueError(
                "Found duplicate view paths under controller management, ensure definitions are unique",
                "\n".join(
                    f"  {view}: {controller}"
                    for view, controllers in duplicate_views
                    for controller in controllers
                ),
            )

        # Validation 2: Ensure that the paths actually exist
        for controller_definition in self.app.controllers:
            controller = controller_definition.controller
            view_path = self.view_root.get_controller_view_path(controller)
            if not view_path.exists():
                raise ValueError(
                    f"View path {view_path} does not exist, ensure it is created before running the server"
                )

    def get_all_root_views(self) -> list[ManagedViewPath]:
        """
        The self.view_root variable is the root of the current user project. We may have other
        "view roots" that store view for plugins.

        This function inspects the controller path definitions and collects all of the
        unique root view paths. The returned ManagedViewPaths are all copied and set to
        share the same package root as the user project.

        """
        # Find the view roots
        view_roots = {self.view_root.copy()}
        for controller_definition in self.app.controllers:
            view_path = controller_definition.controller.view_path
            if isinstance(view_path, ManagedViewPath):
                view_roots.add(view_path.get_root_link().copy())

        # All the view roots should have the same package root
        for view_root in view_roots:
            view_root.package_root_link = self.view_root.package_root_link

        return list(view_roots)

    def get_render_local_state(self, controller: ControllerBase):
        """
        Returns the local type name for the render model. Scoped for use
        within the controller's view directory.

        :returns ReturnModel
        """
        render_metadata = get_function_metadata(controller.render)
        render_model = render_metadata.get_render_model()

        if not render_model:
            raise ValueError(
                f"Controller {controller} does not have a render model defined"
            )

        return camelize(render_model.__name__)

    def openapi_from_controller(self, controller_definition: ControllerDefinition):
        """
        Small hack to get the full path to the root of the server. By default the controller just
        has the path relative to the controller API.

        """
        root_router = APIRouter()
        root_router.include_router(
            controller_definition.router, prefix=controller_definition.url_prefix
        )
        return self.app.generate_openapi(routes=root_router.routes)

    @property
    def openapi_action_specs(self):
        """
        Cache the OpenAPI specs for all side-effects. Render components
        are defined differently. We internally cache this for all stages that require it.

        """
        if not hasattr(self, "_openapi_action_specs"):
            self._openapi_specs: dict[ControllerBase, dict[Any, Any]] = {}

            for controller_definition in self.app.controllers:
                controller = controller_definition.controller
                self._openapi_specs[controller] = self.openapi_from_controller(
                    controller_definition
                )

        return self._openapi_specs

    @property
    def openapi_render_specs(self):
        """
        Get the raw spec for all the render attributes.

        If the return model for a render function is "None", the response spec will
        include {controller: {spec: None}} be so clients can separate undefined controllers from
        defined controllers with no return model.

        """
        if not hasattr(self, "_openapi_render_specs"):
            self._openapi_render_specs: dict[ControllerBase, RenderSpec] = {}

            for controller_definition in self.app.controllers:
                controller = controller_definition.controller

                render_metadata = get_function_metadata(controller.render)
                render_model = render_metadata.get_render_model()

                spec = (
                    self.openapi_schema_converter.get_model_json_schema(render_model)
                    if render_model
                    else None
                )
                self._openapi_render_specs[controller] = RenderSpec(
                    url=controller.url,
                    view_path=str(controller.view_path),
                    spec=spec,
                )

        return self._openapi_render_specs



from .fields import (
    FunctionActionType as FunctionActionType,
    FunctionMetadata as FunctionMetadata,
    fuse_metadata_to_response_typehint as fuse_metadata_to_response_typehint,
    get_function_metadata as get_function_metadata,
    init_function_metadata as init_function_metadata,
)
from .passthrough import passthrough as passthrough
from .sideeffect import sideeffect as sideeffect



from click import command
from mountaineer.cli import handle_runserver, handle_watch


@command()
def runserver():
    handle_runserver(
        package="ci_webapp",
        webservice="ci_webapp.main:app",
        webcontroller="ci_webapp.app:controller",
        port=5006,
        subscribe_to_mountaineer=True,
    )


@command()
def watch():
    handle_watch(
        package="ci_webapp",
        webcontroller="ci_webapp.app:controller",
        subscribe_to_mountaineer=True,
    )



import importlib.metadata
import os
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Flag, auto
from pathlib import Path
from threading import Timer
from typing import Any, Callable

from click import secho
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from mountaineer.paths import resolve_package_path


class CallbackType(Flag):
    CREATED = auto()
    MODIFIED = auto()
    DELETED = auto()


@dataclass
class CallbackEvent:
    action: CallbackType
    path: Path


@dataclass
class CallbackMetadata:
    # Since events can be debounced, we need to send all events that occurred
    # in the batch.
    events: list[CallbackEvent]


@dataclass
class CallbackDefinition:
    action: CallbackType
    callback: Callable[[CallbackMetadata], None]


class ChangeEventHandler(FileSystemEventHandler):
    def __init__(
        self,
        callbacks: list[CallbackDefinition],
        ignore_list=["__pycache__", "_ssr", "_static", "_server"],
        ignore_hidden=True,
        debounce_interval=0.1,
    ):
        """
        :param debounce_interval: Seconds to wait for more events. Will only send one event per batched
        interval to avoid saturating clients with one action that results in many files.

        """
        super().__init__()
        self.callbacks = callbacks
        self.ignore_changes = False
        self.ignore_list = ignore_list
        self.ignore_hidden = ignore_hidden
        self.debounce_interval = debounce_interval
        self.debounce_timer: Timer | None = None
        self.pending_events: list[CallbackEvent] = []

    def on_modified(self, event):
        super().on_modified(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File modified: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.MODIFIED, Path(event.src_path))

    def on_created(self, event):
        super().on_created(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File created: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.CREATED, Path(event.src_path))

    def on_deleted(self, event):
        super().on_deleted(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File deleted: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.DELETED, Path(event.src_path))

    def _debounce(self, action: CallbackType, path: Path):
        if self.debounce_timer is not None:
            self.debounce_timer.cancel()

        self.pending_events.append(CallbackEvent(action=action, path=path))

        self.debounce_timer = Timer(
            self.debounce_interval,
            self.handle_callbacks,
        )
        self.debounce_timer.start()

    def handle_callbacks(self):
        """
        Runs all callbacks for the given action. Since callbacks are allowed to make
        modifications to the filesystem, we temporarily disable the event handler to avoid
        infinite loops.

        """
        self.ignore_changes = True

        for callback in self.callbacks:
            valid_events = [
                event
                for event in self.pending_events
                if event.action in callback.action
            ]
            if valid_events:
                callback.callback(CallbackMetadata(events=valid_events))

        self.pending_events = []
        self.ignore_changes = False

    def should_ignore_path(self, path: Path):
        if self.ignore_changes:
            return True

        path_components = set(str(path).split("/"))

        # Check for any nested hidden directories and ignored directories
        if self.ignore_hidden and any(
            (component for component in path_components if component.startswith("."))
        ):
            return True
        elif path_components & set(self.ignore_list) != set():
            return True
        return False


class WatchdogLockError(Exception):
    def __init__(self, lock_path: Path):
        super().__init__(
            f"Watch lock file exists, another process may be running. If you're sure this is not the case, run:\n"
            f"`$ rm {lock_path}`",
        )
        self.lock_path = lock_path


class PackageWatchdog:
    def __init__(
        self,
        main_package: str,
        dependent_packages: list[str],
        callbacks: list[CallbackDefinition] | None = None,
        run_on_bootup: bool = False,
    ):
        """
        :param run_on_bootup: Typically, we will only notify callback if there has been
            a change to the filesystem. If this is set to True, we will run all callbacks
            on bootup as well.
        """
        self.main_package = main_package
        self.packages = [main_package] + dependent_packages
        self.paths: list[str] = []
        self.callbacks: list[CallbackDefinition] = callbacks or []
        self.run_on_bootup = run_on_bootup

        self.event_handler: ChangeEventHandler | None = None
        self.observer: Any | None = None

        self.check_packages_installed()
        self.get_package_paths()

    def start_watching(self):
        with self.acquire_watchdog_lock():
            if self.run_on_bootup:
                for callback_definition in self.callbacks:
                    callback_definition.callback(CallbackMetadata(events=[]))

            self.event_handler = ChangeEventHandler(callbacks=self.callbacks)
            self.observer = Observer()

            for path in self.paths:
                secho(f"Watching {path}", fg="green")
                if os.path.isdir(path):
                    self.observer.schedule(self.event_handler, path, recursive=True)
                else:
                    self.observer.schedule(
                        self.event_handler, os.path.dirname(path), recursive=False
                    )

            self.observer.start()

            try:
                secho("Starting observer...")
                self.observer.join()
            except KeyboardInterrupt:
                self.observer.stop()
                self.observer.join()

    def check_packages_installed(self):
        for package in self.packages:
            try:
                importlib.metadata.version(package)
            except importlib.metadata.PackageNotFoundError:
                raise ValueError(
                    f"Package '{package}' is not installed in the current environment"
                )

    def get_package_paths(self):
        paths: list[str] = []
        for package in self.packages:
            package_path = resolve_package_path(package)
            paths.append(str(package_path))
        self.paths = self.merge_paths(paths)

    @contextmanager
    def acquire_watchdog_lock(self):
        """
        We only want one watchdog running at a time or otherwise we risk stepping
        on each other or having infinitely looping file change notifications.

        """
        package_path = resolve_package_path(self.main_package)

        lock_path = (Path(str(package_path)) / ".watchdog.lock").absolute()
        if lock_path.exists():
            raise WatchdogLockError(lock_path=lock_path)

        try:
            # Create the lock - this caller should now have exclusive access to the watchdog
            lock_path.touch()
            yield
        finally:
            # Remove the lock
            lock_path.unlink()

    def merge_paths(self, raw_paths: list[str]):
        """
        If one path is a subdirectory of another, we only want to watch the parent
        directory. This function merges the paths to avoid duplicate watchers.

        """
        paths = [Path(path).resolve() for path in raw_paths]

        # Parents should come before their subdirectories
        paths.sort(key=lambda path: len(path.parts))

        merged: list[Path] = []

        for path in paths:
            # Check if the current path is a subdirectory of any path in the merged list
            if not any(path.is_relative_to(parent) for parent in merged):
                merged.append(path)

        # Convert Path objects back to strings
        return [str(path) for path in merged]



from contextlib import asynccontextmanager
from pathlib import Path
from unittest.mock import patch

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from pydantic import BaseModel

from mountaineer.actions import passthrough
from mountaineer.app import AppController
from mountaineer.client_builder.openapi import OpenAPIDefinition
from mountaineer.config import ConfigBase
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException


def test_requires_render_return_value():
    """
    The AppController is in charge of validating our render return value. Since renders are not
    decorated, the best place to validate is during a mount.

    """

    class TestControllerWithoutRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self):
            return None

    class TestControllerWithRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self) -> None:
            return None

    app = AppController(view_root=Path(""))
    with pytest.raises(ValueError, match="must have a return type annotation"):
        app.register(TestControllerWithoutRenderMarkup())

    app.register(TestControllerWithRenderMarkup())


def test_generate_openapi():
    class ExampleSubModel(BaseModel):
        sub_value: str

    class ExampleException(APIException):
        status_code = 401
        invalid_reason: str
        sub_model: ExampleSubModel

    class ExampleController(ControllerBase):
        url = "/example"
        view_path = "/example.tsx"

        def render(self) -> None:
            pass

        @passthrough(exception_models=[ExampleException])
        def test_exception_action(self) -> None:
            pass

    app = AppController(view_root=Path(""))
    app.register(ExampleController())
    openapi_spec = app.generate_openapi()
    openapi_definition = OpenAPIDefinition(**openapi_spec)

    assert openapi_definition.components.schemas.keys() == {
        "TestExceptionActionResponse",
        "ExampleException",
        "ExampleSubModel",
    }


def test_format_exception_model():
    class ExampleException(APIException):
        status_code = 401
        value: str

    app = AppController(view_root=Path(""))
    formatted_exception = app._format_exception_model(ExampleException)

    assert formatted_exception.status_code == 401
    assert formatted_exception.schema_name == "ExampleException"
    assert (
        formatted_exception.schema_name_long
        == "mountaineer.__tests__.test_app.ExampleException"
    )
    assert set(formatted_exception.schema_value["required"]) == {
        "value",
        # Inherited from the superclass
        "status_code",
        "detail",
        "headers",
    }


def test_handle_conflicting_exception_names():
    class Obj1(APIException):
        status_code = 401
        value: str

    class Obj2(APIException):
        status_code = 404
        value: str

    # Since we don't actually want to define these fake
    # modules we just override it at runtime
    Obj1.InternalModel.__name__ = "ExampleException"
    Obj1.InternalModel.__module__ = "mountaineer.__tests__.test_1"
    Obj2.InternalModel.__name__ = "ExampleException"
    Obj2.InternalModel.__module__ = "mountaineer.__tests__.test_2"

    class ExampleController(ControllerBase):
        url = "/example"
        view_path = "/example.tsx"

        def render(self) -> None:
            pass

        @passthrough(exception_models=[Obj1, Obj2])
        def test_exception_action(self) -> None:
            pass

    app = AppController(view_root=Path(""))
    app.register(ExampleController())

    openapi_spec = app.generate_openapi()
    openapi_definition = OpenAPIDefinition(**openapi_spec)

    assert openapi_definition.components.schemas.keys() == {
        "TestExceptionActionResponse",
        "mountaineer.__tests__.test_1.ExampleException",
        "mountaineer.__tests__.test_2.ExampleException",
    }


def test_inherit_parent_spec():
    """
    Ensure we can sniff client functions from the current class
    and the superclasses.

    """

    class ExampleException(APIException):
        status_code = 404
        value: str

    class ParentController(ControllerBase):
        url = "/parent"
        view_path = "/parent.tsx"

        def render(self) -> None:
            pass

        @passthrough(exception_models=[ExampleException])
        def parent_function(self) -> None:
            pass

    class ChildController(ParentController):
        url = "/child"
        view_path = "/child.tsx"

        def render(self) -> None:
            pass

        @passthrough
        def client_function(self) -> None:
            pass

    parent_controller = ParentController()
    child_controller = ChildController()

    app = AppController(view_root=Path(""))
    app.register(parent_controller)
    app.register(child_controller)

    openapi_spec = app.generate_openapi()
    openapi_definition = OpenAPIDefinition(**openapi_spec)

    # Test that we inherited the parent function
    assert (
        "404"
        in openapi_definition.paths["/internal/api/parent_controller/parent_function"]
        .actions[0]
        .responses
    )
    assert (
        "404"
        in openapi_definition.paths["/internal/api/child_controller/parent_function"]
        .actions[0]
        .responses
    )

    # Test that the controller definitions remain separate
    assert parent_controller.definition
    assert child_controller.definition

    parent_routes = parent_controller.definition.render_router.routes
    child_routes = child_controller.definition.render_router.routes

    assert len(parent_routes) == 1
    assert parent_routes[0].path == "/parent"  # type: ignore
    assert len(child_routes) == 1
    assert child_routes[0].path == "/child"  # type: ignore


def test_update_ref_path():
    app = AppController(view_root=Path(""))

    # Test both dictionaries and lists
    fixed_schema = app._update_ref_path(
        {
            "components": {
                "schemas": {
                    "ExampleSubModel": {
                        "properties": {
                            "sub_value": {
                                "$ref": "#/defs/ExampleSubModel",
                            },
                            "list_values": {
                                "type": "array",
                                "items": [
                                    {
                                        "$ref": "#/defs/ExampleSubModel",
                                    }
                                ],
                            },
                        },
                    }
                }
            }
        }
    )

    assert isinstance(fixed_schema, dict)
    properties = fixed_schema["components"]["schemas"]["ExampleSubModel"]["properties"]
    assert properties["sub_value"]["$ref"] == "#/components/schemas/ExampleSubModel"
    assert (
        properties["list_values"]["items"][0]["$ref"]
        == "#/components/schemas/ExampleSubModel"
    )


def test_view_root_from_config(tmp_path: Path):
    class MockConfig(ConfigBase):
        PACKAGE: str | None = "test_webapp"

    # Simulate a package with a views directory
    (tmp_path / "views").mkdir()

    with patch("mountaineer.app.resolve_package_path") as mock_resolve_package_path:
        mock_resolve_package_path.return_value = tmp_path

        app = AppController(config=MockConfig())
        assert app.view_root == tmp_path / "views"

        assert mock_resolve_package_path.call_count == 1
        assert mock_resolve_package_path.call_args[0] == ("test_webapp",)


def test_passthrough_fastapi_args():
    did_run_lifespan = False

    @asynccontextmanager
    async def app_lifespan(app: FastAPI):
        nonlocal did_run_lifespan
        did_run_lifespan = True
        yield

    app = AppController(view_root=Path(""), fastapi_args=dict(lifespan=app_lifespan))

    with TestClient(app.app):
        assert did_run_lifespan


def test_unique_controller_names():
    def make_controller(unique_url: str):
        class ExampleController(ControllerBase):
            url = unique_url
            view_path = unique_url

            def render(self) -> None:
                pass

        return ExampleController

    app = AppController(view_root=Path(""))
    app.register(make_controller("/example")())

    with pytest.raises(ValueError, match="already registered"):
        app.register(make_controller("/example2")())



from uuid import UUID

from fastapi import Request
from mountaineer import ControllerBase, Metadata, RenderBase


class DetailRender(RenderBase):
    client_ip: str


class DetailController(ControllerBase):
    url = "/detail/{detail_id}/"
    view_path = "/app/detail/page.tsx"

    def __init__(self):
        super().__init__()

    def render(self, detail_id: UUID, request: Request) -> DetailRender:
        return DetailRender(
            client_ip=request.client.host if request.client else "unknown",
            metadata=Metadata(title=f"Detail: {detail_id}"),
        )



"""
We specify packages both as python dependencies and venv packages depending
on the local linking strategy. This script ensures that both dependency lists
align so we don't end up out of sync with one install pipeline.

"""

from dataclasses import dataclass
from sys import stdout

import toml

IGNORE_PACKAGE_NAMES = {"python"}


@dataclass(frozen=True, eq=True)
class Package:
    name: str
    extras: frozenset


def parse_poetry_dependencies(data) -> set[Package]:
    deps = set()
    for key, value in data.items():
        if key in IGNORE_PACKAGE_NAMES:
            continue
        package_name = key.split("[")[0]
        extras: frozenset[str] = frozenset()
        if isinstance(value, dict) and "version" in value and "extras" in value:
            extras = frozenset(value["extras"])
        deps.add(Package(name=package_name, extras=extras))
    return deps


def parse_project_dependencies(data) -> set[Package]:
    deps = set()
    for dep in data:
        parts = dep.split("[")
        package_name = parts[0]
        if package_name in IGNORE_PACKAGE_NAMES:
            continue
        extras = frozenset(parts[1][:-1].split(",")) if len(parts) > 1 else frozenset()
        deps.add(Package(name=package_name, extras=extras))
    return deps


def compare_dependencies(poetry_deps: set[Package], project_deps: set[Package]):
    missing_in_project = poetry_deps - project_deps
    missing_in_poetry = project_deps - poetry_deps

    if missing_in_project or missing_in_poetry:
        if missing_in_project:
            stdout.write(f"Missing in project dependencies: {missing_in_project}")
        if missing_in_poetry:
            stdout.write(f"Missing in Poetry dependencies: {missing_in_poetry}")
        exit(1)
    else:
        stdout.write("All dependencies match!")


if __name__ == "__main__":
    with open("pyproject.toml", "r") as toml_file:
        pyproject_data = toml.load(toml_file)

    poetry_deps = parse_poetry_dependencies(
        pyproject_data["tool"]["poetry"]["dependencies"]
    )
    project_deps = parse_project_dependencies(pyproject_data["project"]["dependencies"])

    compare_dependencies(poetry_deps, project_deps)



from typing import Type, TypeVar

from pydantic_settings import BaseSettings

from mountaineer.config import get_config

T = TypeVar("T", bound=BaseSettings)


def get_config_with_type(required_type: Type[T]):
    """
    For use in dependency injection. Will retrieve your
    registered ConfigBase from the global registry and ensure
    that it conforms to the requested type.

    ```python
    async def render(
        self,
        config: AppConfig = Depends(CoreDependencies.get_config_with_type(AppConfig))
    ):
        ...
    ```

    :param required_type: Assert the config class implements this interface. Allows
        for flexible use of multiple BaseSettings to manage different parts of your
        application code.

    """

    def internal_dependency() -> T:
        config = get_config()
        if not isinstance(config, required_type):
            raise TypeError(
                f"Expected config to inherit from {required_type}, {type(config)} is not a valid subclass"
            )
        return config

    return internal_dependency



import asyncio
import socket
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from importlib import import_module
from importlib.metadata import distributions
from multiprocessing import Event, Process, Queue, get_start_method, set_start_method
from multiprocessing.queues import Queue as QueueType
from pathlib import Path
from signal import SIGINT, signal
from tempfile import mkdtemp
from threading import Thread
from time import sleep, time
from traceback import format_exception
from typing import Any, Callable, MutableMapping

from click import secho
from fastapi import Request

from mountaineer.app import AppController
from mountaineer.client_builder.builder import ClientBuilder
from mountaineer.controllers.exception_controller import ExceptionController
from mountaineer.js_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER
from mountaineer.watch import (
    CallbackDefinition,
    CallbackMetadata,
    CallbackType,
    PackageWatchdog,
)
from mountaineer.watch_server import WatcherWebservice
from mountaineer.webservice import UvicornThread


@dataclass
class IsolatedBuildConfig:
    webcontroller: str

    # When builds are completed, a notification will be sent from the subprocess->main process
    # via this channel
    notification_channel: QueueType | None = None

    # If the main process needs to rebuild client js, this flag will open a channel
    # to the subprocess to rebuild the client js
    allow_js_reloads: bool = True

    # Optional arguments to inherit a global cache from the main process
    build_cache: Path | None = None
    build_state: MutableMapping[Any, Any] | None = None


@dataclass
class IsolatedRunserverConfig:
    entrypoint: str
    port: int
    live_reload_port: int


class IsolatedEnvProcess(Process):
    """
    We need a fully separate process for our runserver and watch, so we're able to re-import
    all of the dependent files when there are changes.

    """

    def __init__(
        self,
        build_config: IsolatedBuildConfig,
        runserver_config: IsolatedRunserverConfig | None = None,
    ):
        """
        build_state: State that was originally initialized in the main process, before
        we spawned this worker. This can be used to inherit settings or multiprocessing-safe
        objects that will stick around for the duration of the process lifecycle inbetween-watchers.

        """
        super().__init__()

        self.build_config = build_config
        self.runserver_config = runserver_config
        self.close_signal = Event()
        self.rebuild_channel: QueueType[bool | None] | None = (
            Queue() if build_config.allow_js_reloads else None
        )

    def run(self):
        LOGGER.debug(
            f"Starting isolated environment process with\nbuild_config: {self.build_config}\nrunserver_config: {self.runserver_config}"
        )

        app_controller = import_from_string(self.build_config.webcontroller)
        if not isinstance(app_controller, AppController):
            raise ValueError(
                f"Expected {self.build_config.webcontroller} to be an instance of AppController"
            )

        # Mount our exceptions controller, since we'll need these artifacts built
        # as part of the JS build phase
        self.exception_controller = ExceptionController()
        app_controller.register(self.exception_controller)
        app_controller.app.exception_handler(Exception)(self.handle_dev_exception)

        # Finish the build before we start the server since the server launch is going to sniff
        # for the built artifacts
        if self.build_config is not None:
            # Inject our tmp directory instead of the directories for the already registered
            # build components
            if self.build_config.build_cache:
                for builder in app_controller.builders:
                    builder.tmp_dir = self.build_config.build_cache
            if self.build_config.build_state:
                for builder in app_controller.builders:
                    builder.global_state = self.build_config.build_state

            self.run_build(app_controller)

            # If the client passed a rebuild channel, we'll listen for rebuild requests
            if self.build_config.allow_js_reloads:
                self.listen_for_rebuilds(app_controller)

        if self.runserver_config is not None:
            thread = UvicornThread(
                app=app_controller.app,
                port=self.runserver_config.port,
            )
            thread.start()
            try:
                self.close_signal.wait()
            except KeyboardInterrupt:
                pass
            thread.stop()
            thread.join()

        LOGGER.debug("IsolatedEnvProcess finished")

    def rebuild_js(self):
        LOGGER.debug("JS-Only rebuild started")
        if self.rebuild_channel is not None:
            self.rebuild_channel.put(True)
        else:
            raise ValueError("No rebuild channel was provided")

    def alert_notification_channel(self):
        """
        Alerts the notification channel of a build update, once the server
        comes back online. Before this the client might refresh and get a blank
        page because the server hasn't yet booted.
        """

        def wait_for_server():
            # No need to do anything if we don't have a notification channel
            if self.runserver_config is None:
                return
            if self.build_config.notification_channel is None:
                return

            # Loop until there is something bound to the runserver_config.port
            start = time()
            LOGGER.debug("Waiting for server to come online")
            while True:
                try:
                    with socket.create_connection(
                        ("localhost", self.runserver_config.port)
                    ):
                        break
                except ConnectionRefusedError:
                    sleep(0.1)
            LOGGER.debug(f"Server took {time() - start:.2f} seconds to come online")

            # Buffer to make sure the server is fully booted
            sleep(0.5)

            if self.build_config.notification_channel:
                self.build_config.notification_channel.put(True)

        alert_thread = Thread(target=wait_for_server)
        alert_thread.start()

    def listen_for_rebuilds(self, app_controller: AppController):
        """
        If clients place an object into the rebuild channel, our background thread
        will pick up on these updates and cause a JS-only reload. This only works if the
        application controller's logic hasn't changed, since we use the global one
        that was previously created in our isolated process.

        """

        def wait_for_rebuild():
            if not self.rebuild_channel:
                raise ValueError("No rebuild channel was provided")
            while True:
                rebuild = self.rebuild_channel.get()
                if rebuild is None:
                    break
                self.run_build(app_controller)

        LOGGER.debug("Will launch rebuild thread")
        rebuild_thread = Thread(target=wait_for_rebuild)
        rebuild_thread.start()

    def run_build(self, app_controller: AppController):
        secho("Starting build...", fg="yellow")
        start = time()
        js_compiler = ClientBuilder(
            app_controller,
            live_reload_port=(
                self.runserver_config.live_reload_port
                if self.runserver_config
                else None
            ),
            build_cache=self.build_config.build_cache,
        )
        try:
            js_compiler.build()
            secho(f"Build finished in {time() - start:.2f} seconds", fg="green")

            # Completed successfully
            app_controller.build_exception = None

            self.alert_notification_channel()
        except BuildProcessException as e:
            secho(f"Build failed: {e}", fg="red")
            app_controller.build_exception = e

    def stop(self, hard_timeout: float = 5.0):
        """
        Client-side stop method to shut down the running process.
        """
        # If we've already stopped, don't try to stop again
        if not self.is_alive():
            return

        if self.rebuild_channel is not None:
            self.rebuild_channel.put(None)

        if self.runserver_config is not None:
            self.close_signal.set()

        # Try to give the process time to shut down gracefully
        while self.is_alive() and hard_timeout > 0:
            self.join(1)
            hard_timeout -= 1

        if hard_timeout == 0:
            secho(
                f"Server shutdown reached hard timeout deadline: {self.is_alive()}",
                fg="red",
            )

        # As a last resort we send a hard termination signal
        if self.is_alive():
            self.terminate()

    async def handle_dev_exception(self, request: Request, exc: Exception):
        # If we're receiving a GET request, show the exception. Otherwise fall back
        # on the normal REST handlers
        if request.method == "GET":
            response = await self.exception_controller._generate_html(
                global_metadata=None,
                exception=str(exc),
                stack="".join(format_exception(exc)),
            )
            response.status_code = 500
            return response
        else:
            raise exc


def handle_watch(
    *,
    package: str,
    webcontroller: str,
    subscribe_to_mountaineer: bool = False,
):
    """
    Watch the file directory and rebuild auto-generated files.

    :param package: Ex. "ci_webapp"
    :param webcontroller: Ex. "ci_webapp.app:controller"
    :param subscribe_to_mountaineer:
        If True, will subscribe the local build server to changes in
        the `mountaineer` package. This is useful when doing concurrent
        development in `mountaineer` and a client package. Rarely
        used otherwise.

    """
    update_multiprocessing_settings()

    current_process: IsolatedEnvProcess | None = None

    # The global cache will let us keep cache files warm across
    # different builds
    global_build_cache = Path(mkdtemp())

    def update_build(
        metadata: CallbackMetadata, global_state: MutableMapping[Any, Any]
    ):
        nonlocal current_process

        # JS-Only build needed
        if all(is_view_update(event.path) for event in metadata.events):
            if current_process is not None:
                current_process.rebuild_js()
                return

        if current_process is not None:
            # Stop the current process if it's running
            current_process.stop()

        current_process = IsolatedEnvProcess(
            build_config=IsolatedBuildConfig(
                webcontroller=webcontroller,
                build_cache=global_build_cache,
                build_state=global_state,
            ),
        )
        current_process.start()

    with init_global_state(webcontroller) as global_state:
        watchdog = build_common_watchdog(
            package,
            partial(update_build, global_state=global_state),
            subscribe_to_mountaineer=subscribe_to_mountaineer,
        )
        watchdog.start_watching()


def handle_runserver(
    *,
    package: str,
    webservice: str,
    webcontroller: str,
    port: int,
    subscribe_to_mountaineer: bool = False,
):
    """
    :param package: Ex. "ci_webapp"
    :param webservice: Ex. "ci_webapp.app:app"
    :param webcontroller: Ex. "ci_webapp.app:controller"
    :param port: Desired port for the webapp while running locally
    :param subscribe_to_mountaineer: See `handle_watch` for more details.

    """
    update_multiprocessing_settings()

    current_process: IsolatedEnvProcess | None = None

    # The global cache will let us keep cache files warm across
    # different builds
    global_build_cache = Path(mkdtemp())

    # Start the webservice - it should persist for the lifetime of the
    # runserver, so a single websocket frontend can be notified across
    # multiple builds
    watcher_webservice = WatcherWebservice()
    watcher_webservice.start()

    def update_build(metadata: CallbackMetadata, global_state: dict[Any, Any]):
        nonlocal current_process

        # JS-Only build needed
        if all(is_view_update(event.path) for event in metadata.events):
            if current_process is not None:
                current_process.rebuild_js()
                return

        if current_process is not None:
            # Stop the current process if it's running
            current_process.stop()

        current_process = IsolatedEnvProcess(
            runserver_config=IsolatedRunserverConfig(
                entrypoint=webservice,
                port=port,
                live_reload_port=watcher_webservice.port,
            ),
            build_config=IsolatedBuildConfig(
                webcontroller=webcontroller,
                notification_channel=watcher_webservice.notification_queue,
                build_cache=global_build_cache,
                build_state=global_state,
            ),
        )
        current_process.start()

    # Install a signal handler to catch SIGINT and try to
    # shut down gracefully
    def graceful_shutdown(signum, frame):
        if current_process is not None:
            current_process.stop()
        if watcher_webservice is not None:
            watcher_webservice.stop()
        secho("Services shutdown, now exiting...", fg="yellow")
        exit(0)

    signal(SIGINT, graceful_shutdown)

    with init_global_state(webcontroller) as global_state:
        watchdog = build_common_watchdog(
            package,
            partial(update_build, global_state=global_state),
            subscribe_to_mountaineer=subscribe_to_mountaineer,
        )
        watchdog.start_watching()


def handle_build(
    *,
    webcontroller: str,
):
    """
    Handle a one-off build. Most often used in production CI pipelines.

    """
    app_controller = import_from_string(webcontroller)
    js_compiler = ClientBuilder(
        app_controller,
        live_reload_port=None,
    )
    start = time()
    js_compiler.build()
    secho(f"Build finished in {time() - start:.2f} seconds", fg="green")


def update_multiprocessing_settings():
    """
    fork() is still the default on Linux, and can result in stalls with our asyncio
    event loops. For consistency and expected behavior, try to switch to spawn()
    if it's not already enabled.

    """
    if (spawn_method := get_start_method()) and spawn_method != "spawn":
        LOGGER.warning(
            f"The watch command should be run with the spawn start method, but it's currently set to {spawn_method}",
        )

        try:
            set_start_method("spawn", force=True)
            LOGGER.info("Start method is now set to 'spawn'.")
        except RuntimeError as e:
            # We might catch errors if set_start_method('spawn') is called
            # in an inappropriate context or after multiprocessing has started, which
            # is not allowed.
            LOGGER.error(f"Cannot change the start method after it has been used: {e}")


def import_from_string(import_string: str):
    """
    Given a string to the package (like "ci_webapp.app:controller") import the
    actual variable
    """
    module_name, attribute_name = import_string.split(":")
    module = import_module(module_name)
    return getattr(module, attribute_name)


def find_packages_with_prefix(prefix: str):
    """
    Find and return a list of all installed package names that start with the given prefix.

    """
    return [
        dist.metadata["Name"]
        for dist in distributions()
        if dist.metadata["Name"].startswith(prefix)
    ]


def is_view_update(path: Path):
    """
    Determines if the file change is a view update. This assumes
    the user subscribes to our "views" convention.

    """
    return any(part == "views" for part in path.parts)


def build_common_watchdog(
    client_package: str,
    callback: Callable[[CallbackMetadata], None],
    subscribe_to_mountaineer: bool,
):
    """
    Useful creation class to build a watchdog the common client class
    and our internal package.

    :param subscribe_to_mountaineer: If True, we'll also subscribe to the mountaineer package
    changes in the local environment. This is helpful for local development of the core
    package concurrent with a downstream client application.

    """
    dependent_packages: list[str] = []
    if subscribe_to_mountaineer:
        # Found mountaineer core and mountaineer external dependencies
        dependent_packages = find_packages_with_prefix("mountaineer")
        LOGGER.debug(
            f"Subscribing to changes in local mountaineer packages: {dependent_packages}"
        )

    return PackageWatchdog(
        client_package,
        dependent_packages=dependent_packages,
        callbacks=[
            CallbackDefinition(
                CallbackType.CREATED | CallbackType.MODIFIED,
                callback,
            )
        ],
        # We want to generate a build on the first load
        run_on_bootup=True,
    )


@contextmanager
def init_global_state(webcontroller: str):
    """
    Initialize global state: signal to each builder that they can
    initialize global state before the fork.

    """
    global_state: dict[Any, Any] = {}

    app_controller = import_from_string(webcontroller)

    if not isinstance(app_controller, AppController):
        raise ValueError(f"Unknown app controller: {app_controller}")

    async def entrypoint():
        await asyncio.gather(
            *[builder.init_state(global_state) for builder in app_controller.builders]
        )

    asyncio.run(entrypoint())

    yield global_state



from typing import Any

import pytest
from sqlalchemy.pool import AsyncAdaptedQueuePool, NullPool

from mountaineer.database.config import DatabaseConfig, PoolType
from mountaineer.database.dependencies.core import get_db


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "pool_type, pool_class",
    [
        (PoolType.NULL, NullPool),
        (PoolType.FIXED_PROCESS, AsyncAdaptedQueuePool),
    ],
)
async def test_get_db_null_pool(
    pool_type: PoolType,
    pool_class: Any,
):
    config = DatabaseConfig(
        POSTGRES_HOST="localhost",
        POSTGRES_USER="mock_user",
        POSTGRES_PASSWORD="mock_password",
        POSTGRES_DB="mock_db",
        DATABASE_POOL_TYPE=pool_type,
    )

    db_engine = await get_db(config=config)
    assert isinstance(db_engine.pool, pool_class)



from pathlib import Path

import pytest

from mountaineer.controller import ControllerBase
from mountaineer.render import (
    LinkAttribute,
    MetaAttribute,
    Metadata,
    RenderBase,
    ScriptAttribute,
    ThemeColorMeta,
    ViewportMeta,
)


class StubRenderBase(RenderBase):
    pass


class StubController(ControllerBase):
    view_path = "/page.tsx"

    def render(self):
        return StubRenderBase()


@pytest.mark.parametrize(
    "metadata, expected_tags",
    [
        # Title
        (
            Metadata(
                title="MyTitle",
            ),
            ["<title>MyTitle</title>"],
        ),
        # Multiple links
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet1.css",
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet2.css",
                    ),
                ],
            ),
            [
                '<link rel="stylesheet" href="/stylesheet1.css" />',
                '<link rel="stylesheet" href="/stylesheet2.css" />',
            ],
        ),
        # Mixed meta tags
        (
            Metadata(
                metas=[
                    ThemeColorMeta(
                        color="#000000",
                    ),
                    ViewportMeta(
                        width="device-width",
                        initial_scale=1,
                        maximum_scale=2,
                        user_scalable=False,
                    ),
                ]
            ),
            [
                '<meta name="theme-color" content="#000000" />',
                '<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=2.0, user-scalable=no" />',
            ],
        ),
        # Script tags
        (
            Metadata(
                scripts=[
                    ScriptAttribute(
                        src="/script1.js",
                    ),
                    ScriptAttribute(
                        src="/script2.js",
                        asynchronous=True,
                    ),
                    ScriptAttribute(
                        src="/script3.js",
                        defer=True,
                    ),
                    ScriptAttribute(
                        src="/script4.js",
                        optional_attributes={"test-attr": "test-value"},
                    ),
                ],
            ),
            [
                '<script src="/script1.js"></script>',
                '<script src="/script2.js" async></script>',
                '<script src="/script3.js" defer></script>',
                '<script src="/script4.js" test-attr="test-value"></script>',
            ],
        ),
    ],
)
def test_build_header(metadata: Metadata, expected_tags: list[str]):
    controller = StubController()
    assert controller._build_header(metadata) == expected_tags


COMPLEX_METADATA = Metadata(
    title="MyTitle",
    links=[
        LinkAttribute(
            rel="stylesheet",
            href="/stylesheet1.css",
        )
    ],
    metas=[
        MetaAttribute(
            name="theme-color",
            content="#000000",
        )
    ],
    scripts=[
        ScriptAttribute(
            src="/script1.js",
        )
    ],
)


@pytest.mark.parametrize(
    "metadatas, expected_metadata",
    [
        # A complex metadata definition should always echo its own definition
        ([COMPLEX_METADATA], COMPLEX_METADATA),
        # We shouldn't end up with duplicates
        (
            [
                COMPLEX_METADATA,
                COMPLEX_METADATA,
            ],
            COMPLEX_METADATA,
        ),
        # Test a simple merge of two different values on the same property
        (
            [
                Metadata(
                    links=[
                        LinkAttribute(
                            rel="stylesheet",
                            href="/stylesheet1.css",
                        )
                    ]
                ),
                Metadata(
                    links=[
                        LinkAttribute(
                            rel="stylesheet",
                            href="/stylesheet2.css",
                        )
                    ]
                ),
            ],
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet1.css",
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet2.css",
                    ),
                ]
            ),
        ),
        # The first specified header should win.
        (
            [
                Metadata(title="Primary"),
                Metadata(title="Secondary"),
            ],
            Metadata(title="Primary"),
        ),
    ],
)
def test_merge_metadatas(metadatas: list[Metadata], expected_metadata: Metadata):
    controller = StubController()
    assert controller._merge_metadatas(metadatas) == expected_metadata


def test_resolve_paths(tmp_path: Path):
    view_base = tmp_path / "views"
    ssr_base = view_base / "_ssr"
    static_base = view_base / "_static"

    controller = StubController()
    assert not controller.resolve_paths(view_base)

    # Now create an actual view path that we can sniff
    # This will get further in the pipeline but still won't be valid
    # because we don't have any of the global script files
    view_base.mkdir()
    assert not controller.resolve_paths(view_base)

    # Now we create the SSR script file
    ssr_base.mkdir()
    (ssr_base / "stub_controller.js").touch()
    (ssr_base / "stub_controller.js.map").touch()
    assert not controller.resolve_paths(view_base)

    # Finally, create the static script file
    # Our hash has to be exactly 32 digits to match the regex
    static_base.mkdir()
    random_hash = "b5ecd0c4405374100d6ef93088b86898"
    (static_base / f"stub_controller-{random_hash}.js").touch()
    (static_base / f"stub_controller-{random_hash}.js.map").touch()
    assert controller.resolve_paths(view_base)

    # Now ensure that the paths are correctly set
    assert controller.view_base_path == view_base
    assert controller.ssr_path == ssr_base / "stub_controller.js"
    assert controller.bundled_scripts == [f"stub_controller-{random_hash}.js"]



import pytest_asyncio

from mountaineer.database.dependencies.core import unregister_global_engine


@pytest_asyncio.fixture(autouse=True)
async def unregister_globals():
    await unregister_global_engine()



from contextlib import asynccontextmanager
from functools import partial, wraps
from inspect import Parameter, isawaitable, signature
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    ParamSpec,
    Type,
    TypeVar,
    overload,
)
from urllib.parse import urlparse

from fastapi import Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from starlette.routing import Match

from mountaineer.actions.fields import (
    FunctionActionType,
    ResponseModelType,
    extract_response_model_from_signature,
    handle_explicit_responses,
    init_function_metadata,
)
from mountaineer.cropper import crop_function_for_return_keys
from mountaineer.dependencies import get_function_dependencies
from mountaineer.exceptions import APIException
from mountaineer.render import FieldClassDefinition

if TYPE_CHECKING:
    from mountaineer.controller import ControllerBase

P = ParamSpec("P")
R = TypeVar("R", bound=BaseModel | JSONResponse | None)


@overload
def sideeffect(
    *,
    reload: tuple[Any, ...] | None = None,
    response_model: Type[BaseModel] | None = None,  # deprecated
    exception_models: list[Type[APIException]] | None = None,
    experimental_render_reload: bool | None = None,
) -> Callable[[Callable[P, R | Coroutine[Any, Any, R]]], Callable[P, Awaitable[R]]]:
    ...


@overload
def sideeffect(
    func: Callable[P, R | Coroutine[Any, Any, R]],
) -> Callable[P, Awaitable[R]]:
    ...


def sideeffect(*args, **kwargs):  # type: ignore
    """
    Mark a function as causing a sideeffect to the data. This will force a reload of the full (or partial) server state
    and sync these changes down to the client page.

    Like passthroughs, @sideeffect accepts return values of None, BaseModel, or a JSONResponse if you need full flexibility
    on return headers and content structure. Unlike @passthrough, it does not allow you to provide a non-JSON response since we need
    to internally merge it with render() sideeffect update.

    :param exception_models: List of APIException subclasses that this function is known
        to throw. These will be parsed and available to frontend clients.
    :type exception_models: list[Type[APIException]] | None

    :param reload:

        If provided, will ONLY reload these fields on the client side. By default will reload all fields. Otherwise, why
        specify a sideeffect at all? Note that even if this is provided, we will still regenerate a fully full state on the server
        as if render() is called again. This parameter only controls the data that is streamed back to the client in order to help
        reduce bandwidth of data that won't be changed.

    :type reload: tuple[FieldClassDefinition, ...] | None

    :param experimental_render_reload:

        Experimental options. Disabled by default.<br/><br/>

        If True, will attempt to only execute the logic in render() that is required to calculate your
        `reload` parameters. Other logic will be short-circuited. If your render function has significant computation for other
        properties this can be a significant performance improvement. However, it is experimental and may not work in all cases.

    :type experimental_render_reload: bool

    :return: The response model to use for this endpoint. If a BaseModel is not provided (you pass
        a dictionary or a SQLModel object ofr instance), we will try to convert the response object
        into the proper JSON response based on your typehint.
    :rtype: BaseModel | None | fastapi.JSONResponse

    """

    def decorator_with_args(
        reload: tuple[FieldClassDefinition, ...] | None = None,
        response_model: Type[BaseModel] | None = None,
        exception_models: list[Type[APIException]] | None = None,
        experimental_render_reload: bool = False,
    ):
        def wrapper(func: Callable):
            passthrough_model, response_type = extract_response_model_from_signature(
                func, response_model
            )

            if response_type == ResponseModelType.ITERATOR_RESPONSE:
                raise ValueError(
                    "Sideeffect functions cannot return an iterator response. Use a normal response model instead."
                )

            original_sig = signature(func)
            function_needs_request = "request" in original_sig.parameters

            # Must be delayed until we actually have a self reference
            # Keep a dictionary versus a single value, so we are able to support one
            # controller definition (and therefore a single @sideeffect decorator) being
            # subclassed multiple times
            render_fns: dict[Any, Callable] = {}

            @wraps(func)
            async def inner(self: "ControllerBase", *func_args, **func_kwargs):
                # Delay
                nonlocal render_fns
                render_fn = render_fns.get(self)
                if not render_fn:
                    if experimental_render_reload and reload:
                        render_fn = partial(
                            crop_function_for_return_keys(
                                self.render, keys=[field.key for field in reload]
                            ),
                            self,
                        )
                    else:
                        render_fn = self.render
                    render_fns[self] = render_fn

                # This shouldn't occur - but is necessary for typehinting
                if not render_fn:
                    raise ValueError(
                        "Unable to compute a valid render function for sideeffect"
                    )

                # Check if the original function expects a 'request' parameter
                request = func_kwargs.pop("request")
                if not request:
                    raise ValueError(
                        "Sideeffect function must have a 'request' parameter"
                    )

                if function_needs_request:
                    func_kwargs["request"] = request

                passthrough_values = func(self, *func_args, **func_kwargs)

                # If the original function is async, we now have an awaitable task
                if isawaitable(passthrough_values):
                    passthrough_values = await passthrough_values

                # We need to get the original function signature, and then call it with the request
                async with get_render_parameters(self, request) as values:
                    # Some render functions rely on the URL of the page to make different logic
                    # For this we rely on the Referrer header that is sent on the fetch(). Note that this
                    # referrer can be spoofed, so it assumes that the endpoint also internally validates
                    # the caller has correct permissions to access the data.
                    server_data = render_fn(**values)
                    if isawaitable(server_data):
                        server_data = await server_data

                    return handle_explicit_responses(
                        dict(
                            sideeffect=server_data,
                            passthrough=passthrough_values,
                        )
                    )

            # Update the signature of 'inner' to include 'request: Request'
            # We need to modify this to conform to the request parameters that are sniffed
            # when the component is mounted
            # https://github.com/tiangolo/fastapi/blob/a235d93002b925b0d2d7aa650b7ab6d7bb4b24dd/fastapi/dependencies/utils.py#L250
            sig = signature(inner)
            parameters = list(sig.parameters.values())
            if "request" not in sig.parameters:
                request_param = Parameter(
                    "request", Parameter.POSITIONAL_OR_KEYWORD, annotation=Request
                )
                parameters.insert(1, request_param)  # Insert after 'self'
            new_sig = sig.replace(parameters=parameters)
            inner.__wrapped__.__signature__ = new_sig  # type: ignore

            metadata = init_function_metadata(inner, FunctionActionType.SIDEEFFECT)
            metadata.reload_states = reload
            metadata.passthrough_model = passthrough_model
            metadata.exception_models = exception_models
            metadata.media_type = None  # Use the default json response type
            return inner

        return wrapper

    if args and callable(args[0]):
        # It's used as @sideeffect without arguments
        func = args[0]
        return decorator_with_args()(func)
    else:
        # It's used as @sideeffect(xyz=2) with arguments
        return decorator_with_args(**kwargs)


@asynccontextmanager
async def get_render_parameters(
    controller: "ControllerBase",
    request: Request,
):
    """
    render() components are allowed to have all of the same dependency injections
    that a normal endpoint does. This function parses the render function signature in
    the same way that FastAPI/Starlette do, so we're able to pretend as if a new request
    is coming into the view endpoint.

    NOTE: We exclude calls to background tasks, since these are rarely intended for
    automatic calls to the rendering due to side-effects.

    """
    # Create a synethic request object that we would use to access the core
    # html. This will be passed through the dependency resolution pipeline so to
    # render() it's indistinguishable from a real request and therefore will render
    # in the same way.
    # The referrer should capture the page that they're actually on
    referer = request.headers.get("referer")
    view_request = Request(
        {
            "type": request.scope["type"],
            "path": urlparse(referer or controller.url).path,
            "headers": request.headers.raw,
            "http_version": request.scope["http_version"],
            "method": "GET",
            "scheme": request.scope["scheme"],
            "client": request.scope["client"],
            "server": request.scope["server"],
        }
    )

    if not controller.definition:
        raise RuntimeError(
            "Controller definition is not set. This might indicate you're calling a"
            " sideeffect from outside of an AppController context."
        )

    # Follow starlette's original logic to resolve routes, since this provides us the necessary
    # metadata about URL paths. Unlike in the general-purpose URL resolution case, however,
    # we already know which route should be resolved so we can shortcut having to
    # match non-relevant paths.
    # https://github.com/encode/starlette/blob/5c43dde0ec0917673bb280bcd7ab0c37b78061b7/starlette/routing.py#L544
    for route in controller.definition.render_router.routes:
        match, child_scope = route.matches(view_request.scope)
        if match != Match.FULL:
            raise RuntimeError(
                f"Route {route} did not match ({match}) {view_request.scope}"
            )
        view_request.scope = {
            **view_request.scope,
            "path_params": {},
            "query_string": "",
            **child_scope,
        }

    try:
        async with get_function_dependencies(
            callable=controller.render, url=controller.url, request=view_request
        ) as values:
            yield values
    except RuntimeError as e:
        raise RuntimeError(
            f"Error occurred while resolving dependencies for render(): {controller}: {e}"
        ) from e



from collections import defaultdict
from typing import Type, cast

from pydantic import BaseModel
from sqlalchemy import (
    UniqueConstraint,
    inspect,
)
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql.schema import Table

from mountaineer.database.sqlmodel import SQLModel
from mountaineer.logging import LOGGER


class FieldSchema(BaseModel):
    name: str
    type: str
    primary_key: bool = False
    nullable: bool = True
    foreign_key: str | None = None
    unique: bool = False


class TableSchema(BaseModel):
    name: str
    fields: dict[str, FieldSchema]


class DatabaseValidator:
    """
    Validate that the database schemas are aligned with our ORM defined models. This does a live lookup
    against the current table definitions, so you'll have to pass an active Engine to the constructor.
    Internally we'll use this engine throughout the validation lifecycle.

    """

    def __init__(self, engine):
        self.engine = engine

    def validate_database_alignment(self, models: list[Type[SQLModel]]):
        """
        Returns the discrepancies between the current database schema and the expected schema, if any.

        """
        current_schema = self.get_current_schema()
        expected_schema = self.parse_model_definitions(models)
        discrepancies = self.compare_schemas(current_schema, expected_schema)
        return {
            table_key: list(found_issues)
            for table_key, found_issues in discrepancies.items()
        }

    def parse_model_definitions(
        self, models: list[Type[SQLModel]]
    ) -> dict[str, TableSchema]:
        """
        Parse the Python model for the current definitions
        """
        model_schemas: dict[str, TableSchema] = {}
        for model in models:
            fields: dict[str, FieldSchema] = {}
            table = cast(Table, model.__table__)  # type: ignore
            primary_keys = {pk.name for pk in table.primary_key}
            unique_constraints = {
                col.name
                for constraint in table.constraints
                if isinstance(constraint, UniqueConstraint)
                for col in constraint.columns
            }
            foreign_keys = {
                fk.parent.name: fk.column.table.name + "(" + fk.column.name + ")"
                for fk in table.foreign_keys
            }

            for column in table.columns:
                field_schema = FieldSchema(
                    name=column.name,
                    type=str(column.type),
                    primary_key=column.name in primary_keys,
                    nullable=column.nullable if column.nullable is not None else True,
                    foreign_key=foreign_keys.get(column.name),
                    unique=column.name in unique_constraints,
                )
                fields[column.name] = field_schema
            table_name = model.__tablename__
            if not isinstance(table_name, str):
                raise ValueError(
                    f"Table name for model {model} is not a string: {table_name}"
                )
            model_schemas[table_name] = TableSchema(name=table_name, fields=fields)
        return model_schemas

    def get_current_schema(self) -> dict[str, TableSchema]:
        """
        Parse the database for the current table schema
        """
        inspector: Inspector = inspect(self.engine)
        tables_schema: dict[str, TableSchema] = {}

        for table_name in inspector.get_table_names():
            columns_info = inspector.get_columns(table_name)
            pk_constraint = inspector.get_pk_constraint(table_name)
            fk_constraints = inspector.get_foreign_keys(table_name)
            indexes = inspector.get_indexes(table_name)

            # Mapping of primary key columns
            primary_keys = set(pk_constraint.get("constrained_columns", []))
            # Mapping for unique columns (simplified, considering only unique indexes, not unique constraints)
            unique_columns = {
                index["column_names"][0]
                for index in indexes
                if index["unique"] and len(index["column_names"]) == 1
            }

            table_schema: dict[str, FieldSchema] = {}
            for column_info in columns_info:
                column_name = column_info["name"]
                foreign_key = None
                for fk_constraint in fk_constraints:
                    if column_name in fk_constraint.get("constrained_columns", []):
                        # Simplified representation of foreign key as 'referenced_table(referenced_column)'
                        ref_table = fk_constraint["referred_table"]
                        ref_column = (
                            fk_constraint["referred_columns"][0]
                            if fk_constraint["referred_columns"]
                            else None
                        )
                        foreign_key = (
                            f"{ref_table}({ref_column})" if ref_column else ref_table
                        )
                        break

                field_schema = FieldSchema(
                    name=column_name,
                    type=column_info["type"].__str__(),
                    primary_key=column_name in primary_keys,
                    nullable=column_info["nullable"],
                    foreign_key=foreign_key,
                    unique=column_name in unique_columns,
                )

                table_schema[column_name] = field_schema

            tables_schema[table_name] = TableSchema(
                name=table_name, fields=table_schema
            )

        return tables_schema

    def compare_schemas(
        self, db_schema: dict[str, TableSchema], model_schema: dict[str, TableSchema]
    ):
        discrepancies = defaultdict(list)

        # Check for tables missing in the database
        for table in model_schema:
            if table not in db_schema:
                discrepancies[table].append("Missing table in database")
                continue  # No need to compare fields if the table is missing

        for table, model_fields in model_schema.items():
            LOGGER.info(f"Comparing {table}")
            db_fields = db_schema.get(table)

            for field_name, model_field in model_fields.fields.items():
                LOGGER.info(f"Comparing {table}.{field_name}")
                db_field = db_fields.fields.get(field_name) if db_fields else None

                # Field missing in the database
                if not db_field:
                    discrepancies[f"{table}.{field_name}"].append(
                        "Missing field in database"
                    )
                    continue

                # Compare field type
                if model_field.type != db_field.type:
                    discrepancies[f"{table}.{field_name}"].append(
                        f"Type mismatch (Model: {model_field.type}, DB: {db_field.type})"
                    )

                # Compare primary key status
                if model_field.primary_key != db_field.primary_key:
                    discrepancies[f"{table}.{field_name}"].append(
                        f"Primary key mismatch (Model: {model_field.primary_key}, DB: {db_field.primary_key})"
                    )

                # Compare nullable status
                if model_field.nullable != db_field.nullable:
                    discrepancies[f"{table}.{field_name}"].append(
                        f"Nullable mismatch (Model: {model_field.nullable}, DB: {db_field.nullable})"
                    )

                # Compare unique constraint status
                if model_field.unique != db_field.unique:
                    discrepancies[f"{table}.{field_name}"].append(
                        f"Unique constraint mismatch (Model: {model_field.unique}, DB: {db_field.unique})"
                    )

                # Compare foreign keys
                if model_field.foreign_key != db_field.foreign_key:
                    discrepancies[f"{table}.{field_name}"].append(
                        f"Foreign key mismatch (Model: {model_field.foreign_key}, DB: {db_field.foreign_key})"
                    )

        # Check for fields in DB not present in model
        for table, db_fields in db_schema.items():
            if table in model_schema:  # Only check if table is expected to exist
                for field_name in db_fields.fields:
                    if field_name not in model_schema[table].fields:
                        discrepancies[f"{table}.{field_name}"].append(
                            "Extra field in database"
                        )

        return discrepancies



"""
Collection of utilities to "crop" python functions to only execute a sub-AST worth
of logic that's required to produce the output value.

"""
import ast
import inspect
from copy import copy
from textwrap import dedent
from typing import Any, Callable

from pydantic import BaseModel

from mountaineer.logging import LOGGER

DependencyGraphType = dict[str, set[str]]


class FunctionCropException(Exception):
    pass


def var_to_synthetic_var(var: str):
    return f"return_synthetic_{var}"


class SyntheticVarInserter(ast.NodeTransformer):
    """
    Extract any expressions in our return statement:

    return {
        "key1": value1 + value2,
    }

    And turn into synthetic variables that we can then easily traverse via a dependency graph:

    synthetic_key1 = value1 + value2
    return {
        "key1": synthetic_key1,
    }

    """

    def __init__(self, known_pydantic_models: list[str]):
        self.known_pydantic_models = known_pydantic_models

    def visit_Return(self, node):
        new_stmts = []

        if isinstance(node.value, ast.Dict):  # Direct dictionary returns
            for i, (key, value) in enumerate(zip(node.value.keys, node.value.values)):
                key_str = key.value if isinstance(key, ast.Constant) else None
                if key_str:
                    assign, synthetic_var_name = self.create_synthetic_assign(
                        key_str, value
                    )
                    new_stmts.append(assign)
                    node.value.values[i] = ast.Name(
                        id=synthetic_var_name, ctx=ast.Load()
                    )

        elif (
            isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Name)
            and (
                node.value.func.id == "dict"
                or node.value.func.id in self.known_pydantic_models
            )
        ):
            # Handle Pydantic model returns
            # Handle dict() function returns
            for i, keyword in enumerate(node.value.keywords):
                key_str = keyword.arg
                if key_str:
                    assign, synthetic_var_name = self.create_synthetic_assign(
                        key_str, keyword.value
                    )
                    new_stmts.append(assign)
                    node.value.keywords[i].value = ast.Name(
                        id=synthetic_var_name, ctx=ast.Load()
                    )

        else:
            raise FunctionCropException(
                "Unknown return type, can't auto-crop function logic."
            )

        return new_stmts + [node] if new_stmts else node

    def create_synthetic_assign(self, key: str, value: ast.expr):
        """
        Create a synthetic variable assignment statement.
        """
        synthetic_var_name = var_to_synthetic_var(key)
        return ast.Assign(
            targets=[ast.Name(id=synthetic_var_name, ctx=ast.Store())], value=value
        ), synthetic_var_name


class DependencyGraphCreator(ast.NodeVisitor):
    """
    Analyze a function's AST to determine which variables depend on which other variables

    """

    def __init__(self):
        self.graph: DependencyGraphType = {}
        self.current_deps: list[str] = []

    def visit_Assign(self, node):
        # Collect dependencies for the value being assigned
        self.current_deps = []
        self.visit(node.value)
        for target in node.targets:
            if isinstance(target, ast.Name):
                # Update the graph with dependencies for this assignment
                self.graph[target.id] = set(self.current_deps)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            # A variable is being used here, add it to the current dependencies
            self.current_deps.append(node.id)

    def visit_Call(self, node):
        # For simplicity, assume function calls are side-effect free and only depend on
        # their arguments
        for arg in node.args:
            self.visit(arg)


class ASTReducer(ast.NodeTransformer):
    """
    Extract a sub-AST from a given AST, based on a list of needed variables. This will keep
    all variables within `needed_vars` and all non-assigned expressions.

    """

    def __init__(self, needed_vars: set[str], known_pydantic_models: list[str]):
        super().__init__()
        self.needed_vars = needed_vars
        self.known_pydantic_models = known_pydantic_models

    def visit_FunctionDef(self, node: ast.FunctionDef):
        return self.visit_function_common(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        return self.visit_function_common(node)

    def visit_If(self, node):
        # Process the body and orelse parts of the If statement
        node.body = [stmt for stmt in node.body if self.is_needed(stmt)]
        node.orelse = [stmt for stmt in node.orelse if self.is_needed(stmt)]

        # If both body and orelse become empty, exclude the If statement
        if not node.body and not node.orelse:
            return None
        return node

    def visit_function_common(self, node: ast.AsyncFunctionDef | ast.FunctionDef):
        new_body: list[ast.stmt] = []
        for stmt in node.body:
            if isinstance(stmt, ast.If):
                if stmt := self.visit_If(stmt) and stmt:
                    new_body.append(stmt)
            elif self.is_needed(stmt):
                new_body.append(stmt)
            elif isinstance(stmt, ast.Return):
                modified_return = self.modify_return_stmt(stmt)
                if modified_return:
                    new_body.append(modified_return)
        node.body = new_body
        return node

    def is_needed(self, stmt: ast.stmt):
        if isinstance(stmt, ast.Assign):
            return any(
                target.id in self.needed_vars
                for target in stmt.targets
                if isinstance(target, ast.Name)
            )
        elif isinstance(stmt, ast.Expr):
            # For now assume that we need all expressions
            return True
        elif isinstance(stmt, ast.If):
            # Check if the If statement contains needed variables in its body or orelse
            return any(self.is_needed(sub_stmt) for sub_stmt in stmt.body + stmt.orelse)
        return False

    def modify_return_stmt(self, stmt: ast.Return):
        # Prepare the dictionary keys and values based on the model's fields (arguments to the model)
        new_keys: list[ast.Constant] = []
        new_values: list[ast.expr] = []

        if (
            isinstance(stmt.value, ast.Call)
            and hasattr(stmt.value.func, "id")
            and (
                stmt.value.func.id == "dict"  # type: ignore
                or stmt.value.func.id in self.known_pydantic_models  # type: ignore
            )
        ):
            for key, value in zip(
                [ast.Constant(value=arg.arg) for arg in stmt.value.keywords],
                [arg.value for arg in stmt.value.keywords],
            ):
                if isinstance(value, ast.Name) and value.id in self.needed_vars:
                    new_keys.append(key)
                    new_values.append(value)
        elif isinstance(stmt.value, ast.Dict):
            for dict_key, dict_value in zip(stmt.value.keys, stmt.value.values):
                if (
                    isinstance(dict_key, ast.Constant)
                    and isinstance(dict_value, ast.Name)
                    and dict_value.id in self.needed_vars
                ):
                    new_keys.append(dict_key)
                    new_values.append(dict_value)
        return ast.Return(value=ast.Dict(keys=new_keys, values=new_values))


def reduce_function_to_keys(
    func_ast: ast.Module,
    graph: DependencyGraphType,
    target_keys: list[str],
    known_pydantic_models: list[str],
):
    needed_vars = set(target_keys)
    queue = copy(target_keys)

    while queue:
        current_var = queue.pop()
        if current_var in graph:
            dependencies = graph[current_var]
            for dep in dependencies:
                if dep not in needed_vars:
                    needed_vars.add(dep)
                    queue.append(dep)

    reducer = ASTReducer(needed_vars, known_pydantic_models)
    return reducer.visit(func_ast)


def crop_function_for_return_keys(
    func: Callable, keys: list[str], locals: dict[str, Any] | None = None
):
    """
    Performs static analysis on the given function. Expects this function to either return
    a dictionary or a BaseModel.

    Returns a new synthetic function that always returns a dictionary, just with the subset
    of keys that are specified. Will only perform the computation necessary to directly calculate
    these keys, which saves compute for other non-required functions.

    Known limitations:
        - Only one "return" statement is supported. If you have conditional logic that returns different
            response payloads, this cropping will not work. Instead, have one return payload with multiple
            conditional input variables.

    """
    source = inspect.getsource(func)
    dedented_source = dedent(source)
    tree = ast.parse(dedented_source)

    isolated_namespace = func.__globals__.copy()
    if locals:
        isolated_namespace.update(locals)

    known_pydantic_models = [
        model.__name__
        for model in isolated_namespace.values()
        if inspect.isclass(model) and issubclass(model, BaseModel)
    ]

    # The user is referencing object keys in the resulting payload, refactor these
    # into actual variables
    inserter = SyntheticVarInserter(known_pydantic_models)
    tree = inserter.visit(tree)
    keys = [var_to_synthetic_var(var) for var in keys]

    # Create a dependency graph
    creator = DependencyGraphCreator()
    creator.visit(tree)

    LOGGER.debug(f"Parsed function graph: {creator.graph}")

    # Reduce the function based on the dependency graph
    optimized_tree = reduce_function_to_keys(
        tree, creator.graph, keys, known_pydantic_models
    )

    # Fix line numbers and compile
    optimized_tree = ast.fix_missing_locations(optimized_tree)
    code = compile(optimized_tree, filename="<ast>", mode="exec")

    exec(code, isolated_namespace)

    return isolated_namespace[func.__name__]



from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Coroutine, MutableMapping

from pydantic import BaseModel

from mountaineer.controller import ControllerBase
from mountaineer.paths import ManagedViewPath


class ClientBundleMetadata(BaseModel):
    live_reload_port: int | None = None


class ClientBuilderBase(ABC):
    """
    Base class for client builders. When mounted to an AppController, these build plugins
    will be called for every file defined in the view/app directory. It's up to the plugin
    whether to handle the incoming file.

    """

    def __init__(self, tmp_dir: Path | None = None):
        # We keep a tmpdir open for the duration of the build process, so our rust
        # logic can leverage file-based caches for faster builds
        # Note that this tmpdir is shared across all client builders, so it's important
        # that you enforce uniqueness of filenames if you leverage this cache
        self.tmp_dir = tmp_dir if tmp_dir else Path(mkdtemp())
        self.global_state: MutableMapping[Any, Any] | None = None

    async def init_state(self, global_state: MutableMapping[Any, Any]):
        pass

    async def start_build(self):
        pass

    @abstractmethod
    async def handle_file(
        self,
        file_path: ManagedViewPath,
        controller: ControllerBase | None,
        metadata: ClientBundleMetadata,
    ) -> None | Coroutine[Any, Any, None]:
        """
        Only direct controller views are called with (view, controller) inputs. Otherwise we do a
        recursive search of the raw files on disk with controller=None.

        """
        pass

    async def finish_build(self):
        pass



import asyncio
from multiprocessing import Queue
from threading import Thread

from fastapi import FastAPI, WebSocket
from starlette.websockets import WebSocketDisconnect

from mountaineer.io import get_free_port
from mountaineer.logging import LOGGER
from mountaineer.webservice import UvicornThread


class WatcherWebservice:
    """
    A simple webserver to notify frontends about updated builds.

    The WatcherWebservice provides a multiprocessing safe queue that's
    accessible as `notification_queue`. Each time that a process
    wants to update the frontend, it can push a message into the queue.

    """

    def __init__(self, webservice_port: int | None = None):
        self.app = self.build_app()
        self.websockets: list[WebSocket] = []
        self.port = webservice_port or get_free_port()
        self.notification_queue: Queue[bool | None] = Queue()

        self.webservice_thread: UvicornThread | None = None
        self.monitor_build_thread: Thread | None = None

        self.has_started = False

    def build_app(self):
        app = FastAPI()

        @app.get("/")
        def home():
            return {"message": "Hello World"}

        @app.websocket("/build-events")
        async def build_updated(websocket: WebSocket):
            await websocket.accept()
            self.websockets.append(websocket)

            while True:
                # Keep the connection open until the client side disconnects
                try:
                    data = await websocket.receive_text()
                except WebSocketDisconnect:
                    # Expected exception, don't log an error
                    self.websockets.remove(websocket)
                    break
                await websocket.send_text(f"echo: {data}")

        return app

    async def broadcast_listeners(self):
        for ws in self.websockets:
            await ws.send_text("Build updated")
        LOGGER.info("Broadcasted build update to %d listeners", len(self.websockets))

    def monitor_builds(self):
        while True:
            next_obj = self.notification_queue.get()
            if next_obj is None:
                break
            asyncio.run(self.broadcast_listeners())

    def start(self):
        if self.has_started:
            raise Exception("WatcherWebservice has already started")

        self.webservice_thread = UvicornThread(
            app=self.app,
            port=self.port,
            log_level="warning",
        )

        LOGGER.debug("Starting WatcherWebservice on port %d", self.port)
        self.webservice_thread.start()

        self.monitor_build_thread = Thread(target=self.monitor_builds)
        self.monitor_build_thread.start()

        self.has_started = True

    def stop(self):
        if self.webservice_thread is not None:
            self.webservice_thread.stop()
            self.webservice_thread.join()
        if self.monitor_build_thread is not None:
            self.notification_queue.put(None)
            self.monitor_build_thread.join()
        LOGGER.info("WatcherWebservice has stopped")



from mountaineer.database.config import DatabaseConfig


class ExampleDatabaseConfig(DatabaseConfig):
    POSTGRES_HOST: str = "localhost"
    POSTGRES_USER: str = "mountaineer"
    POSTGRES_PASSWORD: str = "raw_password"
    POSTGRES_DB: str = "mountaineer"


def test_allows_default_params():
    """
    Test that the config model should merge the default and the runtime
    provided values, with a preference to runtime values.

    """
    config = ExampleDatabaseConfig(POSTGRES_PASSWORD="password")

    assert config.POSTGRES_PASSWORD == "password"
    assert "password_raw" not in str(config.SQLALCHEMY_DATABASE_URI)



from contextlib import contextmanager
from os import chdir
from pathlib import Path
from re import sub

import pytest

from scripts.update_version import (
    format_cargo_version,
    format_python_version,
    update_version_python,
    update_version_rust,
)


@contextmanager
def change_pwd(new_pwd: Path):
    old_pwd = Path.cwd()
    try:
        chdir(new_pwd)
        yield
    finally:
        chdir(old_pwd)


@pytest.mark.parametrize(
    "input_version,expected",
    [
        ("1.2.3", "1.2.3"),
        ("1.2", "1.2.0"),
        ("1.2.3-alpha.1", "1.2.3-a1"),
        ("1.2.3-beta.2", "1.2.3-b2"),
        ("1.2.3.post1", "1.2.3-post1"),
        ("1.2.3.dev4", "1.2.3-dev4"),
        ("1.2.3-alpha.1.post2.dev3", "1.2.3-a1-post2-dev3"),
    ],
)
def test_format_cargo_version(input_version, expected):
    assert format_cargo_version(input_version) == expected


def test_update_version_rust(tmp_path: Path):
    cargo_text = """
        [package]
        name = "mountaineer"
        # Bumped automatically by CI on a release
        version = "0.1.0"
        edition = "2021"

        [dependencies]
        v8 = "0.89.0"
        deno_core_icudata = "0.73.0"
        """

    cargo_path = tmp_path / "Cargo.toml"
    cargo_path.write_text(cargo_text)

    with change_pwd(tmp_path):
        update_version_rust("0.2.0")

    assert sub(r"\s+", "", cargo_path.read_text()) == sub(
        r"\s+",
        "",
        """
        [package]
        name = "mountaineer"
        version = "0.2.0"
        edition = "2021"

        [dependencies]
        v8 = "0.89.0"
        deno_core_icudata = "0.73.0"
        """,
    )


@pytest.mark.parametrize(
    "input_version,expected",
    [
        ("1.2.3", "1.2.3"),
        ("1.2.3a1", "1.2.3b1"),
        ("1.2.3b2", "1.2.3b2"),
        ("1.2.3rc3", "1.2.3b3"),
        ("1.2.3.post1", "1.2.3.post1"),
        ("1.2.3.dev4", "1.2.3.dev4"),
        ("1.2.3a1.post2.dev3", "1.2.3b1.post2.dev3"),
    ],
)
def test_format_python_version(input_version, expected):
    assert format_python_version(input_version) == expected


def test_update_version_python(tmp_path: Path):
    pyproject_text = """
        [tool.poetry]
        name = "mountaineer"
        version = "0.1.0"
        description = ""
        readme = "README.md"

        [tool.poetry.dependencies]
        uvicorn = { extras = ["standard"], version = "^0.27.0.post1" }
        fastapi = "^0.68.0"
        """

    pyproject_path = tmp_path / "pyproject.toml"
    pyproject_path.write_text(pyproject_text)

    with change_pwd(tmp_path):
        update_version_python("0.2.0")

    assert sub(r"\s+", "", pyproject_path.read_text()) == sub(
        r"\s+",
        "",
        """
        [tool.poetry]
        name = "mountaineer"
        version = "0.2.0"
        description = ""
        readme = "README.md"

        [tool.poetry.dependencies]
        fastapi = "^0.68.0"

        [tool.poetry.dependencies.uvicorn]
        extras = ["standard",]
        version = "^0.27.0.post1"
        """,
    )



import subprocess
from pathlib import Path

from click import secho


def has_npm():
    try:
        subprocess.run(
            ["npm", "--version"],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except subprocess.CalledProcessError:
        return True
    except FileNotFoundError:
        return False


def npm_install(path: Path):
    try:
        subprocess.run(
            ["npm", "install"],
            check=True,
            cwd=str(path),
        )
        return True
    except Exception as e:
        secho(
            f"An unexpected error occurred while installing npm dependencies: {e}",
            fg="red",
        )
        return False


def get_git_user_info():
    """
    Get the global Git user name and email, if git is set up on the system.

    """
    try:
        git_user_name = (
            subprocess.check_output(["git", "config", "--global", "user.name"])
            .strip()
            .decode()
        )
        git_user_email = (
            subprocess.check_output(["git", "config", "--global", "user.email"])
            .strip()
            .decode()
        )
        return git_user_name, git_user_email
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None, None



from enum import StrEnum
from pathlib import Path

from jinja2 import Template
from pydantic import BaseModel


class EditorType(StrEnum):
    VSCODE = "vscode"
    VIM = "vim"
    ZED = "zed"


class ProjectMetadata(BaseModel):
    project_name: str
    author_name: str
    author_email: str
    use_poetry: bool
    use_tailwind: bool
    editor_config: EditorType | None
    project_path: Path

    postgres_password: str = "mysecretpassword"
    postgres_port: int = 5432

    create_stub_files: bool

    # Current version of mountaineer tied to CMA version
    mountaineer_min_version: str

    # Path components to the project's default virtual environment
    # Not set until after the environment is created
    venv_base: str | None = None
    venv_name: str | None = None

    # If specified, will install mountaineer in development mode pointing to a local path
    # This is useful for testing changes to mountaineer itself
    mountaineer_dev_path: Path | None = None


class TemplateOutput(BaseModel):
    content: str
    path: str


def format_template(
    path: Path, base_path: Path, project_metadata: ProjectMetadata
) -> TemplateOutput:
    """
    Takes in a template path (relative to /templates) and returns the formatted
    template contents and the final path of the file.

    We support:
        - Jinja templating within the file
        - Bracket syntax in filenames, like /path/to/[project_name]/file.txt

    """
    if not path.exists():
        raise FileNotFoundError(f"Template file {path} does not exist")

    metadata_variables = project_metadata.model_dump()

    template = Template(path.read_text())
    content = template.render(metadata_variables)

    output_name = str(path.relative_to(base_path))
    for key, value in metadata_variables.items():
        output_name = output_name.replace(f"[{key}]", str(value))

    return TemplateOutput(content=content, path=output_name)
